"""
Builder for asymptote (.asy) files
"""
from .builder import Builder
from .save_temp import SaveTempFile
from .composite_builders import SequentialBuilder


class Asy2pdf(Builder):
    """A builder to convert from asy to pdf."""

    action = "asy -safe -f pdf -o {builder.outfilepath} {builder.infilepaths}"
    available = True
    priority = 1000
    required_execs = ('asy',)

    infilepath_ext = '.asy'
    outfilepath_ext = '.pdf'


class SaveAsyPdf(SequentialBuilder):
    """A builder that renders text to an asy file and converts it to pdf."""

    available = True
    priority = 1000
    infilepath_ext = '.save'
    outfilepath_ext = '.pdf'

    def __init__(self, env, subbuilders=None, use_cache=None, **kwargs):
        # setup the subbuilders
        subbuilders = subbuilders if subbuilders is not None else []

        # Setup a SaveTempFile file to save the Asy file
        save_temp = SaveTempFile(env=env, save_ext='.asy', use_cache=True,
                                 **kwargs)
        subbuilders.append(save_temp)
        self.subbuilder_for_outfilename = save_temp

        # Setup a Asy2pdf builder
        asy2pdf = Asy2pdf(env=env, use_cache=True, **kwargs)
        subbuilders.append(asy2pdf)

        super().__init__(env=env, subbuilders=subbuilders, use_cache=use_cache,
                         **kwargs)


class Asy2svg(SequentialBuilder):
    """A CompositeBuilder for Asy2svg that includes Asy2pdf and Pdf2svg
    builders.
    """

    available = True
    priority = 1000
    infilepath_ext = '.asy'
    outfilepath_ext = '.svg'

    def __init__(self, env, subbuilders=None, target=None, use_cache=None,
                 **kwargs):
        # Setup parameters
        subbuilders = (list(subbuilders) if isinstance(subbuilders, list) or
                       isinstance(subbuilders, tuple) else [])

        # Create the subbuilders
        asy2pdf = Asy2pdf(env=env, target=target, use_cache=True, **kwargs)
        subbuilders.append(asy2pdf)

        pdf2svg_cls = self.find_builder_cls(in_ext='.pdf', out_ext='.svg',
                                            target=target)
        pdf2svg = pdf2svg_cls(env=env, target=target, use_cache=True, **kwargs)
        subbuilders.append(pdf2svg)

        super().__init__(env, subbuilders=subbuilders, use_cache=use_cache,
                         target=target, **kwargs)


class SaveAsySvg(SequentialBuilder):
    """A builder that renders text to an asy file and converts it to svg."""

    available = True
    priority = 1000
    infilepath_ext = '.save'
    outfilepath_ext = '.svg'

    def __init__(self, env, subbuilders=None, use_cache=None, **kwargs):
        # setup the subbuilders
        subbuilders = subbuilders if subbuilders is not None else []

        # Setup a SaveTempFile file to save the Asy file
        save_temp = SaveTempFile(env=env, save_ext='.asy', use_cache=True,
                                 **kwargs)
        subbuilders.append(save_temp)
        self.subbuilder_for_outfilename = save_temp

        # Setup a Asy2pdf builder
        asy2pdf = Asy2svg(env=env, use_cache=True, **kwargs)
        subbuilders.append(asy2pdf)

        super().__init__(env=env, subbuilders=subbuilders, use_cache=use_cache,
                         **kwargs)



"""
Receiver for document load to reset the context and document
"""
from ..signals import document_onload, document_deleted


@document_onload.connect_via(order=100)
def reset_document(document, **kwargs):
    """Reset the context and managers for a document on load."""
    context = document.context or dict()

    # Reset the context
    context.reset()

    return document


@document_deleted.connect_via(order=100)
def delete_document(document, **kwargs):
    """Reset the context and managers for a document on document deletion."""
    context = document.context or dict()

    # Reset the context
    del context
    if document.context is not None:
        delattr(document, 'context')

    return document



"""
Functions to validate arguments
"""


def _raise_error(msg, exception, raise_error=True):
    """When an argument is not valid, raise a ValidationError with the
    given message when the raise_error argument is True. Otherwise return
    False.
    """
    if raise_error:
        raise exception(msg)

    return False


def validate_tuple(arg, length=None, type=None, raise_error=True):
    """Validate or try to convert an arg (argument) into a tuple.

    Parameters
    ----------
    arg : Iterable
        The tuple (or other iterable) to validate.
    length : Optional[Int]
        If specified, the tuple is only valid if it has this length.
    type : Optional[Type]
        If specified, this function will try to convert the items to this type,
        and the tuple will only be valid if its items are of this type.
    raise_error : Optional[Bool]
        If True (default), a ValidationError will be raised if arg doesn't
        validate.

    Returns
    -------
    arg : Union[arg, False]
        Returns the validated tuple.
        If the tuple could not be validated, returns False

    Examples
    --------
    >>> validate_tuple('test')
    Traceback (most recent call last):
        ...
    TypeError: The argument 'test' must not be a string
    >>> validate_tuple(1)
    Traceback (most recent call last):
        ...
    TypeError: The argument '1' must be a tuple
    >>> validate_tuple([1, 2, 3], length=4)
    Traceback (most recent call last):
        ...
    ValueError: The argument tuple '(1, 2, 3)' must be of length '4'
    >>> validate_tuple([1, 2, 3], length=3)
    (1, 2, 3)
    >>> validate_tuple([1.0, 2.0, 3.0], length=3, type=int)
    (1, 2, 3)
    >>> validate_tuple([1.0, 'a', 3.0], length=3, type=int)
    Traceback (most recent call last):
        ...
    ValueError: invalid literal for int() with base 10: 'a'
    """
    if isinstance(arg, str):
        return _raise_error("The argument '{}' must not be a "
                            "string".format(arg),
                            TypeError, raise_error=raise_error)
    if not hasattr(arg, '__iter__'):
        return _raise_error("The argument '{}' must be a tuple".format(arg),
                            TypeError, raise_error=raise_error)
    try:
        arg = tuple(arg)
    except TypeError:
        return _raise_error("The argument '{}' could not be converted to "
                            "a tuple".format(arg), TypeError,
                            raise_error=raise_error)

    if length is not None and len(arg) != length:
        return _raise_error("The argument tuple '{}' must be of length "
                            "'{}'".format(arg, length), ValueError,
                            raise_error=raise_error)

    if type is not None:
        try:
            arg = tuple(type(i) for i in arg)
        except TypeError:
            return _raise_error("The items of tuple '{}' could not be "
                                "converted to type '{}'".format(arg, type),
                                TypeError, raise_error=raise_error)
    return arg



from .load import load_document
from .reset import reset_document, delete_document
from .process_headers import process_headers
from .process_tags import process_tags
from .process_document_label import process_document_label

__all__ = ('load_document', 'reset_document', 'delete_document',
           'process_headers', 'process_tags', 'process_document_label')



"""
Decider classes to evaluate whether a build is needed.
"""
import pathlib

from ...utils.classes import weakattr


class Decision(object):
    """A decision to build.

    The base decision checks to see if the input files and output file
        exist.

    Parameters
    ----------
    parent_decider : :obj:`.builders.decider.Decider`
        The parent decider instance that created this decision.
    """

    parent_decider = weakattr()

    def __init__(self, parent_decider):
        self.parent_decider = parent_decider

    def build_needed(self, inputs, output, reset=False):
        """Determine whether a build is needed.

        Parameters
        ----------
        inputs : List[str, :obj:`.paths.SourcePath`, tuple]
            The input infilepaths, strings and arguments to use in the build.
        output : :obj:`.paths.TargetPath`
            The outfilepath for the built file
        reset : Optional[bool]
            If True, reset cached values in determining whether the build is
            needed.

        Raises
        ------
        MissingInputFiles
            Raise if one or more of the input files are missing.
        """
        # inputs should be a list or tuple with some items in the inputs
        assert ((isinstance(inputs, list) or isinstance(inputs, tuple)) and
                len(inputs) > 0)

        # Test to make sure all of the SourcePath inputs exist
        infiles = [p for p in inputs if isinstance(p, pathlib.Path)]
        if not all(p.exists() for p in infiles):
            # This returns True because a builder may be a subbuilder whose
            # input files aren't available yet.
            return True
        elif isinstance(output, pathlib.Path) and not output.exists():
            return True
        else:
            return False


class Decider(object):
    """A decider to evaluate whether a build is needed."""

    environment = weakattr()
    decision_cls = Decision

    def __init__(self, env):
        self.environment = env

    @property
    def decision(self):
        """Return the Decision class associated with this decider."""
        return self.decision_cls(parent_decider=self)



"""
Tags for figure environments.
"""
from .tag import Tag
from .caption import Caption
from .utils import xhtml_percentwidth, tex_percentwidth
from ..utils.string import strip_multi_newlines
from .. import settings


class BaseFigure(Tag):
    """A base class for a figure tag.

    The BaseFigure initializes figures by adding 'id' attributes to labels in
    the label manager and reorganizing captions to the bottom of the figure.
    """

    # Render the figures as '<span>' elements in xhtml, instead of <figure>
    # elements becase <figure> elements are block level items--i.e. they
    # cannot be placed inside a paragraph.
    html_name = 'figure'
    active = False

    include_paragraphs = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Transfer the label id ('id') to the caption, if available. First,
        # find the caption tag, if available
        captions = [tag for tag in self.flatten(filter_tags=True)
                    if isinstance(tag, Caption)]

        for caption in captions:
            # Transfer the 'id' to the caption (but only the first)
            caption.label_id = self.attributes.pop('id', None)

            # Set the label kind for the caption as a figure caption
            caption.kind = ('caption', 'figure')

            # Make the caption a figcaption tag for html
            caption.html_name = 'figcaption'
            caption.html_class = 'figcaption'

            # Create the label in the label_manager
            caption.create_label()

    def html_fmt(self, attributes=None, method='html', **kwargs):
        attrs = attributes or self.attributes.copy()

        # Set the html class
        if self.html_class is not None:
            if 'class' in attrs:
                attrs['class'] = ' ' + self.html_class
            else:
                attrs['class'] = self.html_class
        return super().html_fmt(attributes=attrs, method=method, **kwargs)


class MarginFigure(BaseFigure):
    """The @marginfig tag"""

    aliases = ('marginfig',)
    html_class = 'marginfig'
    tex_env = 'marginfigure'
    active = True

    def xhtml_fmt(self, attributes=None, **kwargs):
        attributes = attributes or self.attributes.copy()
        attr_name = '{{{}}}type'.format(settings.xhtml_namespace['epub'])
        attributes[attr_name] = 'footnote'
        return super().xhtml_fmt(attributes=attributes, **kwargs)


class Figure(BaseFigure):
    """The @figure/@fig tag"""

    aliases = ('fig',)
    html_class = 'figure'
    tex_env = 'figure'
    active = True


class FullFigure(BaseFigure):
    """The @fullfigure/@ffig tag"""

    aliases = ('ffig', 'fullfig')
    html_class = 'fullfigure'
    tex_env = 'figure*'
    active = True


class Panel(Tag):
    """A panel (sub-figure) for a figure."""

    active = True
    html_name = 'panel'
    tex_env = 'panel'

    def tex_fmt(self, attributes=None, **kwargs):
        attrs = self.attributes.copy() if attributes is None else attributes

        # Format the width
        attrs = tex_percentwidth(attrs, target='.tex', use_positional=True)

        # Raises an error if a width is not present. Strip multiple newlines
        # as these break up side-by-side figures
        env = super().tex_fmt(attributes=attrs, **kwargs)
        return strip_multi_newlines(env).strip()

    def html_fmt(self, attributes=None, method='html', **kwargs):
        attrs = self.attributes.copy() if attributes is None else attributes
        target = '.' + method if not method.startswith('.') else method

        # Format the html classes
        attrs['class'] = 'panel'
        attrs = xhtml_percentwidth(attributes=attrs, target=target)
        return super().html_fmt(attributes=attrs, method=method, **kwargs)



"""Checkers for .tex targets."""

import subprocess

from .checker import Checker
from .types import SoftwareDependency, SoftwareDependencyList
from ..software_dependencies import pdf_deps
from ..utils.list import chunks


class PdfChecker(Checker):
    """Checker for pdf renderings from latex."""

    def __init__(self, pdf_deps=pdf_deps):
        """Initialize checker for pdf rendering dependencies."""
        super().__init__('pdf', *pdf_deps.dependencies)

    def check_classes(self, classes=None):
        """Check the availability of classes.

        Parameters
        ----------
        classes : Union[List[str], Tuple[str]]
            The list of packages to check for availability.
        """
        # Get a listing of 'check_classes' methods
        check_classes_meth_names = [meth for meth in dir(self)
                                    if meth.startswith('check_classes_')]

        # Execute the methods
        for meth_name in check_classes_meth_names:
            meth = getattr(self, meth_name)
            meth(classes)

    def check_kpsewhich(self, listing, ext):
        """Check the availability of latex packages using kpsewhich.

        Parameters
        ----------
        listing : Union[List[str], Tuple[str]]
            Base filenames for latex classes or packages to check.
            ex: 'article'
        ext : str
            The extension to search. ex: '.sty' or '.cls'
        """
        # Make sure that 'kpsewhich' is installed.
        if self.find_executable('kpsewhich') is None:
            return None

        # Evalulate the presence of packages by opening at most 5
        # processes at
        # a time
        for chunk in chunks(listing, 5):
            processes = []

            # Start the processes for kpsewhich
            for item in chunk:
                if isinstance(item, SoftwareDependency):
                    # construct the kpsewhich command
                    args = ('kpsewhich', item.name + ext)

                    # Run the subprocess
                    p = subprocess.Popen(args, stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE,
                                         bufsize=4096, )
                    processes.append((item, p))

                elif isinstance(item, SoftwareDependencyList):
                    self.check_kpsewhich(item.dependencies, ext=ext)

            # Wait for the processes to have terminated
            for item, process in processes:
                # Check that it was succesfully executed
                out, err = process.communicate()
                returncode = process.returncode

                # Assign the package dependency's availability
                item.available = (returncode == 0)  # succesfully found

    def check_packages_kpsewhich(self, packages=None):
        """Check available latex packages (.sty) using kpsewhich.

        Parameters
        ----------
        packages : Union[List[str], Tuple[str]]
                The list of packages to check for availability.
        """
        if packages is not None:
            pass
        elif 'packages' in self:
            packages = self['packages']
        else:
            return None
        self.check_kpsewhich(packages, '.sty')

    def check_fonts(self, fonts=None):
        """Check available of latex fonts (.tfm) using kpsewhich.

        Parameters
        ----------
        fonts : Union[List[str], Tuple[str]]
                The list of fonts to check for availability.
        """
        if fonts is not None:
            pass
        elif 'fonts' in self:
            fonts = self['fonts']
        else:
            return None
        self.check_kpsewhich(fonts, '.tfm')

    def check_classes_kpsewhich(self, classes=None):
        """Check available latex classes (.cls) using kpsewhich.

        Parameters
        ----------
        classes : Union[List[str], Tuple[str]]
            The list of classes to check for availability.
        """
        if classes is not None:
            pass
        elif 'classes' in self:
            classes = self['classes']
        else:
            return None
        self.check_kpsewhich(classes, '.cls')



"""
A request handler for rendering the signals.
"""
from .server import ServerHandler
from ...signals.signals import signals
from ...utils.string import stub


class SignalHandler(ServerHandler):
    """A request handler for the signals"""

    def get(self):
        # Get the signals
        signals_dict = self.signals_to_dict(signal_namespace=signals)

        self.render('signals.html', signals=signals_dict)

    def signals_to_dict(self, signal_namespace):
        """Given a list of processor classes, prepare a list of dicts listing
        the sub-processors
        Parameters
        ----------
        signal_namespace : :obj:`.signals.signals.Namespace`
        """
        signals_list = []

        # Sort the processor_clses
        for count, (name, signal) in enumerate(signal_namespace.items(), 1):
            d = dict()

            # populate the meta data
            d['name'] = name

            # Get the docstring (first line), if available
            if getattr(signal, '__doc__', None) is not None:
                docstring = stub(signal.__doc__)
                d['doc'] = docstring

            # Get the receivers for the signal
            receiver_items = signal.receivers_dict().items()
            receivers = list(enumerate(sorted(receiver_items)))
            receiver_list = d.setdefault('receivers', [])

            for count, (order, receiver) in receivers:
                r = dict()
                r['name'] = receiver.__name__
                r['order'] = order

                if receiver.__doc__ is not None:
                    docstring = stub(receiver.__doc__)
                    r['doc'] = docstring

                receiver_list.append(r)

            signals_list.append(d)

        return signals_list



"""
Tags for note environments.
"""
from .tag import Tag


class Sidenote(Tag):
    """A sidenote tag."""

    aliases = ('marginnote',)

    html_name = 'sidenote'
    tex_cmd = 'marginnote'

    active = True



"""
Utilities for paths.
"""
import pathlib


def rename(path, filename=None, append=None, extension=None):
    """Rename the filename from a path with optional modifiers.

    Parameters
    ----------
    path : Union[str, :obj:`pathlib.Path`]
        The path to rename
    filename : Optional[str]
        The new filename to rename the filename (without the extension)
    append : Optional[str]
        If specified, add the string to the filename
    extension : Optional[str]
        If specified, replace the extension with the given extension

    Returns
    -------
    new_path : :obj:`pathlib.Path`
        The new path with the filename renamed.

    Examples
    --------
    >>> rename('/tmp/ch1_file.png', filename='test')
    PosixPath('/tmp/test.png')
    >>> rename('/tmp/ch1_file.png', append='_crop2')
    PosixPath('/tmp/ch1_file_crop2.png')
    >>> rename('/tmp/ch1_file.png', append='_crop2', extension='ext')
    PosixPath('/tmp/ch1_file_crop2.ext')
    >>> rename('/tmp/ch1.1_file.png', filename='test')
    PosixPath('/tmp/test.png')
    >>> rename('/tmp/ch1.1_file.png', filename='test', append='_crop2')
    PosixPath('/tmp/test_crop2.png')
    >>> rename('/tmp/ch1.1_file.png', filename='test', append='_crop2',
    ...        extension='svg')
    PosixPath('/tmp/test_crop2.svg')
    >>> rename('/tmp/ch1.1_file', extension='html')  # this won't work
    PosixPath('/tmp/ch1.html')
    """
    # Prepare the path argument
    if isinstance(path, str):
        path = pathlib.Path(path)
    parent = path.parent  # directory of path

    # Get the base filename (stem) and workup as necessary
    filename = path.stem if filename is None else filename
    if append:
        filename += append

    # Get the extension and workup as necessary
    extension = path.suffix if extension is None else extension
    extension = '.' + extension if not extension.startswith('.') else extension

    # Return the renamed path
    return parent / (filename + extension)


def find_files(string, context):
    """Determine if filenames for valid files are present in the given
    string.

    Parameters
    ----------
    string : str
        The string that may contain filenames
    context : :obj:`.DocumentContext`
        A document context that contains paths to search.

    Returns
    -------
    filepaths : List[:obj:`pathlib.Path`]
        List of found paths
    """
    assert 'paths' in context and isinstance(context['paths'], list)

    # Setup arguments and return values
    # First remove extra spaces and newlines on the ends.
    string = str(string).strip()
    filepaths = list()

    # Go line-by-line and find valid files. Stop when no more lines can
    # be found
    for line in string.splitlines():
        # Skip empty lines
        if line == "":
            continue

        filepath = find_file(line, context=context, raise_error=False)
        if filepath is not None:
            filepaths.append(filepath)

    return filepaths


def find_file(path, context, raise_error=True):
    """Search for an existing file given the path and, if needed, the 'paths'
    entry from a context.

    Parameters
    ----------
    path : Union[str, :obj:`pathlib.Path`]
        The path stub to search.
    context : :obj:`.context.Context`
        The document's context with an entry of 'paths' to search.
    raise_error : Optional[bool]
        If True (default), raise error if a file isn't found.

    Returns
    -------
    valid_path : :obj:`pathlib.Path`
        A path for a file that exists.

    Raises
    ------
    FileNotFoundError
        Raised if the file could not be found.
    """
    assert 'paths' in context and isinstance(context['paths'], list)

    # Prepare the parameters
    if isinstance(path, str):
        path_strs = path.strip().splitlines()  # Keep only the first line
        path = pathlib.Path(path_strs[0] if path_strs else '')

    # If it's not a path at this stage, return None--no path can be found
    if not isinstance(path, pathlib.Path):
        if raise_error:
            msg = "Could not find file '{}'".format(path)
            raise FileNotFoundError(msg)
        return None

    # See if the file is itself a valid file. This might fail if the string
    # in the path would make a filename that is too long, raising an OSError
    try:
        if path.is_file():
            return path
    except OSError as e:
        if raise_error:
            raise e

    # See if it's an absolute file that doesn't exist
    if path.is_absolute():
        if raise_error:
            msg = "Could not find file '{}'".format(path)
            raise FileNotFoundError(msg)
        return None

    # Otherwise see if the path can be reconstructed
    paths = ['.'] + context['paths']
    for p in paths:
        # Treat SourcePath and TargetPath in a special way to correctly set the
        # subpath
        if hasattr(p, 'use_subpath'):
            subpath = p.subpath / path if p.subpath is not None else path
            new_path = p.use_subpath(subpath)
        else:
            new_path = p / path

        try:
            if new_path.is_file():
                return new_path
        except OSError as e:
            if raise_error:
                raise e

    # I'm out of ideas and places to search! Raise an error
    if raise_error:
        msg = "Could not find file '{}' in the paths: '{}'".format(path, paths)
        raise FileNotFoundError(msg)
    return None



"""
Classes and functions for rendering documents.
"""
from shutil import rmtree
from collections import OrderedDict
import logging
import pathlib

from .document_context import DocumentContext
from . import exceptions, signals
from ..paths import SourcePath
from .. import settings


class Document(object):
    """A base class document rendered from a source file to one or more
    target files.

    Parameters
    ----------
    src_filepath : Union[:obj:`SourcePath <.paths.SourcePath>`, str]
        The path (an absolute path or a path relative to the current directory)
        for the document (markup source) file of this document.
    target_root : Optional[Union[:obj:`TargetPath <.paths.TargetPath>`, str]]
        The path for the rendered target files. Subdirectories for the targets
        will be created. (ex: 'html' 'tex')
        By default, if not specified, the target_root will be one directory
        above the project_root.
    parent_context : Optional[:obj:`DocumentContext <.DocumentContext>`]
        The context of the parent document or default context.
    level : Optional[int]
            The level of document loaded. Sub-documents are loaded at a higher
            level (>1) than the root document.

    Attributes
    ----------
    src_filepath : :obj:`SourcePath <.paths.SourcePath>`
        The filename and path for this document's (markup source) file. This
        file should exist.
        ex: 'src/chapter1/chapter1.dm'
    context : :obj:`DocumentContext <.DocumentContext>`
        A context dict with the values needed to render a target document.
    subdocuments : OrderedDict[:obj:`SourcePath <.paths.SourcePath>`, \
        :obj:`Document <.Document>`]
        An ordered dict with the sub-documents included in this document.
        The keys are src_filepaths, and the values are the sub-documents
        themselves. The documents are ordered according to their placement in
        the document tree (project).

        This document owns the sub-documents and only weak references to these
        documents should be made to these documents. (i.e. when the
        sub-documents dict is cleared, the memory for the document objects
        should be released)
    """

    src_filepath = None
    context = None
    subdocuments = None

    #: The path of a temporary directory created for this document, if needed.
    _temp_dir = None

    #: A flag to determine whether the document was successfully loaded
    _succesfully_loaded = False

    def __init__(self, src_filepath, environment, parent_context=None,
                 level=1):
        logging.debug("Creating document: {}".format(src_filepath))

        # Populate attributes
        self.subdocuments = OrderedDict()
        self._templates = dict()  # FIXME: Remove

        # Process the paths
        project_root = environment.project_root
        target_root = environment.target_root
        if isinstance(src_filepath, SourcePath):
            self.src_filepath = src_filepath
        else:
            src_filepath = pathlib.Path(src_filepath)
            subpath = src_filepath.relative_to(project_root)
            self.src_filepath = SourcePath(project_root=project_root,
                                           subpath=subpath)

        # Create the context
        self.context = DocumentContext(document=self,
                                       project_root=project_root,
                                       target_root=target_root,
                                       environment=environment,
                                       parent_context=parent_context)

        # Read in the document and load sub-documents
        self.load(level=level)

        # Send the 'document_created' signal
        signals.document_created.emit(document=self)

    def __del__(self):
        """Clean up any temp directories no longer in use."""
        # Send the 'document_deleted' signal
        signals.document_deleted.emit(document=self)

        if self._temp_dir is not None:
            rmtree(self._temp_dir, ignore_errors=True)

    def __repr__(self):
        return "Document({})".format(self.src_filepath)

    @property
    def doc_id(self):
        """The unique string identifier (within a project) for the document."""
        # The subpath of the src_filepath is guaranteed to be unique.
        return str(self.src_filepath.subpath)

    @property
    def doc_ids(self):
        """The list of all doc_id for this document and all subdocuments."""
        return [doc.doc_id for doc in self.documents_list(recursive=True)]

    @property
    def title(self):
        """The title for the document."""
        if 'title' in self.context:
            # The title entry could be a string or a tag. If it's a tag, just
            # get the text for the tag.
            title = self.context['title']
            title_str = (title.default_fmt().strip()
                         if hasattr(title, 'default_fmt')
                         else title)
            return title_str
        return str(self.src_filepath.subpath.with_suffix(''))

    @property
    def short(self):
        """The short title for the document."""
        if 'short' in self.context:
            # The short entry could be a string or a tag. If it's a tag, just
            # get the text for the tag.
            short = self.context['short']
            short_str = (short.default_fmt().strip()
                         if hasattr(short, 'default_fmt')
                         else short)
            return short_str
        else:
            return self.title

    @property
    def mtime(self):
        return self.context.get('mtime', None)

    @property
    def project_root(self):
        return self.context.get('project_root', None)

    @property
    def target_root(self):
        return self.context.get('target_root', None)

    @property
    def targets(self):
        """The targets dict.

        There may be more targets in this dict than those listed in the
        context because some targets may be intermediary targets in a cache
        path.

        Returns
        -------
        targets : Dict[str, :obj:`TargetPath <.paths.TargetPath>`]
            The targets are a dict with the target extension as keys
            (ex: '.html') and the value is the target_filepath for that target.
            (ex: 'html/index.html') These paths are target paths.
        """
        return self.context.target_filepaths()

    def target_filepath(self, target):
        """The filepath for the given target extension.

        Parameters
        ----------
        target : str
            The target extension for the target file.
            ex: '.html' or '.tex'

        Returns
        -------
        target_filepath : Union[:obj:`TargetPath <.paths.TargetPath>`, None]
            The target filepath.
        """
        return self.context.target_filepath(target=target)

    @property
    def label_manager(self):
        return self.context.get('label_manager', None)

    # TODO: rename to documents_by_src_filepath
    def documents_dict(self, document=None, only_subdocuments=False,
                       recursive=False):
        """Produce an ordered dict of a document and all its sub-documents.

        Parameters
        ----------
        document : Optional[:obj:`Document <.Document>`]
            The document for which to create the document list.
            If None is specified, this document will be used.
        only_subdocuments : Optional[bool]
            If True, only the sub-documents will be returned (not this
            document or the document specified.)
        recursive : Optional[bool]
            If True, the sub-documents of sub-documents are returned (in order)
            in the ordered dict as well

        Returns
        -------
        document_dict : OrderedDict[:obj:`SourcePath <.paths.SourcePath>`, \
            :obj:`Document <.Document>`]
            An ordered dict of documents. The keys are src_filepaths and the
            values are document objects.
        """
        document = self if document is None else document
        doc_dict = OrderedDict()

        if not only_subdocuments:
            doc_dict[document.src_filepath] = document

        for src_filepath, subdoc in document.subdocuments.items():

            if recursive:
                subdoc_dict = subdoc.documents_dict(only_subdocuments=False,
                                                    recursive=recursive)
                for k, v in subdoc_dict.items():
                    doc_dict[k] = v
            else:
                doc_dict[subdoc.src_filepath] = subdoc

        return doc_dict

    def documents_by_id(self, document=None, only_subdocuments=False,
                        recursive=True):
        """Produce an ordered dict of a document and sub-documents entered by
        doc_id.

        Parameters
        ----------
        document : Optional[:obj:`Document <.Document>`]
            The document for which to create the document dict.
            If None is specified, this document will be used.
        only_subdocuments : Optional[bool]
            If True, only the sub-documents will be returned (not this
            document or the document specified.)
        recursive : Optional[bool]
            If True, the sub-documents of sub-documents are returned (
            in order) in the ordered dict as well

        Returns
        -------
        document_dict : OrderedDict[str, :obj:`Document <.Document>`]
            An ordered dict of documents. The keys are doc_ids
            and the values are document objects.
        """
        doc_dict = self.documents_dict(document=document,
                                       only_subdocuments=only_subdocuments,
                                       recursive=recursive)
        dict_by_id = OrderedDict([(doc.doc_id, doc)
                                  for doc in doc_dict.values()])

        if len(dict_by_id) < len(doc_dict):
            # there are duplicate doc_ids
            msg = ("The project has multiple documents with the same doc_id. "
                   "The doc_id should be unique for each document.")
            raise exceptions.DocumentException(msg)

        return dict_by_id

    def documents_list(self, document=None, only_subdocuments=False,
                       recursive=False):
        """Produce an ordered list of a document and all its sub-documents.

        Parameters
        ----------
        document : Optional[:obj:`Document <.Document>`]
            The document for which to create the document list.
        only_subdocuments : Optional[bool]
            If True, only the sub-documents will be returned (not this root
            document or the document specified.
        recursive : Optional[bool]
            If True, the sub-documents of sub-documents are returned (in order)
            in the list as well

        Returns
        -------
        document_list : List[:obj:`Document <.Document>`]
            An ordered list of documents.
        """
        doc_dict = self.documents_dict(document=document,
                                       only_subdocuments=only_subdocuments,
                                       recursive=recursive)

        return list(doc_dict.values())

    def load_required(self):
        """Evaluate whether a load is required.

        Returns
        -------
        load_required : bool
            True, if a load is required.
            False if a load isn't required.
        """
        # Reload the document if:
        # 1. The document hasn't been successfully loaded.
        if not self._succesfully_loaded:
            logging.debug("Load required for {}: The document was not "
                          "previously loaded successfully.".format(self))
            return True

        # 2. The body attribute hasn't been set yet.
        body_attr = settings.body_attr
        if self.context.get(body_attr, None) is None:
            logging.debug("Load required for {}: The '{}' tag has "
                          "not been loaded yet.".format(self, body_attr))
            return True

        # 3. The mtime for the file is now later than the one stored in the
        #    context--i.e. the user saved the file.
        src_mtime = self.src_filepath.stat().st_mtime
        last_mtime = self.mtime
        if last_mtime is None or src_mtime > last_mtime:
            logging.debug("Load required for {}: The '{}' source file is "
                          "newer than the loaded "
                          "document.".format(self, self.src_filepath.subpath))
            return True

        # 4. This is a subdocument whose context has a parent_context, and the
        #    parent_context has been updated
        parent_context = self.context.parent_context
        parent_mtime = (parent_context.get('mtime')
                        if parent_context is not None else None)
        if parent_mtime is not None and parent_mtime > last_mtime:
            logging.debug("Load required for {}: The parent document is newer "
                          "than this document.".format(self))
            return True

        return False

    def load(self, reload=False, level=1):
        """Load or reload the document into the context.

        Parameters
        ----------
        reload : Optional[bool]
            If True, force the reload of the document.
        level : Optional[int]
            The level of document loaded. Sub-documents are loaded at a higher
            level (>1) than the root document.

        Returns
        -------
        document_loaded : bool
            True, if a sub-document was (re)loaded.
        """
        document_loaded = False
        # Check to make sure the file exists
        if not self.src_filepath.is_file():  # file must exist
            msg = "The source document '{}' must exist."
            raise exceptions.DocumentException(msg.format(self.src_filepath))

        # Load document if a load is required or forced
        if self.load_required() or reload:

            # The document hasn't been loaded yet. Reset the flat
            self._succesfully_loaded = False

            # Check to make sure the file is reasonable
            stat = self.src_filepath.stat()
            filesize = stat.st_size
            if filesize > settings.document_max_size:
                msg = ("The source document '{}' has a file size ({} kB) "
                       "that exceeds the maximum setting size of {} kB.")
                actual_filesize = filesize / 1024
                max_filesize = settings.document_max_size / 1024
                msg = msg.format(self.src_filepath, actual_filesize,
                                 max_filesize)
                raise exceptions.DocumentException(msg)

            # Emit the load signal
            signals.document_onload.emit(document=self, context=self.context)

            # The document has been loaded
            self._succesfully_loaded = True
            document_loaded |= True

        # Load the sub-documents. This is done after processing the string
        # and ast since these may include sub-files, which should be read
        # in before being loaded.
        document_loaded |= self.load_subdocuments(level=level + 1)

        # Emit a signal if this is the root document and a document (or sub-
        # document) was reloaded
        if document_loaded and level == 1:
            root_document = self.context.root_document
            signals.document_tree_updated.emit(root_document=root_document)

        return document_loaded

    @staticmethod
    def _update_mtime(document, mtime=None):
        """Update the mtime of all subdocuments for a given document if the
        document's mtime is later.

        Parameters
        ----------
        document : :obj:`.Document`
            The root document.
        mtime : Optional[float]
            The modification time to update, if newer
        """
        document_mtime = getattr(document, 'mtime')
        mtime = document_mtime if mtime is None else mtime
        subdocuments = getattr(document, 'subdocuments')

        # Update the document's mtime
        if (mtime is not None and document_mtime is not None and
                mtime > document_mtime):
            document.context['mtime'] = mtime

        if subdocuments:
            # Process the subdocuments
            [Document._update_mtime(doc, mtime)
             for doc in subdocuments.values()]

        return mtime

    def load_subdocuments(self, level=1):
        """Load the sub-documents listed in the include entries in the
        context.

        Parameters
        ----------
        level : Optional[int]
            The level of document loaded. Sub-documents are loaded at a higher
            level (>1) than the root document.

        Returns
        -------
        document_loaded : bool
            True, if a sub-document was (re)loaded.
        """
        document_loaded = False

        # Get the root document's src_filepath and the src_filepath for all of
        # its included subdocuments to make sure we do not load documents
        # recursively or twice
        root_document = self.context.get('root_document', None)
        root_document = root_document() if root_document is not None else None

        # Get the build environment for the project
        environment = self.context['environment']

        # Get a dict with all documents in a project
        root_dict = root_document.documents_dict(document=root_document,
                                                 only_subdocuments=False,
                                                 recursive=True)

        # Get a dict with all documents owned by this document. This document
        # can control those documents, but it cannot control other documents in
        # the root_dict. This avoids recursion.
        subs_dict = self.documents_dict(only_subdocuments=False,
                                        recursive=True)

        # Clear the subdocuments ordered dict and add new entries. Old entries
        # will automatically be deleted if the subdocument no longer holds a
        # reference to it.
        self.subdocuments.clear()

        # Retrieve the file paths of included files in the context
        src_filepaths = self.context.includes

        for src_filepath in src_filepaths:
            # If the src_filepath is the same as this document, do nothing, as
            # a document shouldn't have itself as a subdocument
            if src_filepath == self.src_filepath:
                continue

            # If the document is already loaded and controlled by this
            # document, copy it to the subdocuments ordered dict
            elif src_filepath in subs_dict:
                subdoc = subs_dict[src_filepath]
                document_loaded |= subdoc.load(level=level + 1)

                if self.src_filepath not in subdoc.subdocuments:
                    self.subdocuments[src_filepath] = subdoc

            # The document could not be found, at this point. Create it--as
            # long as it's not controlled by another document--i.e. it's not
            # already loaded in root_dict
            elif src_filepath not in root_dict:
                subdoc = Document(src_filepath=src_filepath,
                                  environment=environment,
                                  parent_context=self.context,
                                  level=level + 1)
                self.subdocuments[src_filepath] = subdoc
                document_loaded |= True
            else:
                continue

        # Update the subdocuments mtimes to this document's mtime. The
        # sub-document's mtime must updated so that a load_required is not
        # triggered by a newer parent_context. Even if the subdoc's mtime is
        # updated to a newer version, it should not cause problems in
        # triggering a load_required for the subdoc if the src_filepath is
        # updated because the src_filepath's mtime will be later than the
        # current document's mtime.
        Document._update_mtime(document=self)

        return document_loaded

    def build_needed(self):
        """Evaluate whether a build is required"""
        # Reload document
        self.load()
        return any(signals.document_build_needed.emit(document=self))

    def build(self, complete=True):
        """Run a build of the document and all subdocuments."""
        # Make sure the document (and subdocuments) are loaded
        self.load()

        # Send the 'document_build' signal
        statuses = signals.document_build.emit(document=self,
                                               complete=complete)
        return statuses[0] if len(statuses) == 1 else statuses



"""
Builder to save temporary files
"""
import logging

from .builder import Builder, BuildError
from ..utils.classes import weakattr


class SaveTempFile(Builder):
    """A Builder to save a string to a temporary file

    Parameters
    ----------
    env: :obj:`.builders.Environment`
        The build environment
    parameters, args : Tuple[:obj:`pathlib.Path`, str, tuple, list]
        The input parameters (dependencies), including the string to save to
        the temporary outfilepath, for the build
    outfilepath : Optional[:obj:`.paths.TargetPath`]
        If specified, the path for the output file.
    save_ext : str
        The extension for the saved temp file.
    """

    available = True
    action = 'save'
    priority = 1000
    active_requirements = ('priority',)
    scan_parameters_on_init = False  # Scan after all parameters are loaded

    context = weakattr()

    infilepath_ext = '.save'  # dummy extension for find_builder_cls

    def __init__(self, env, context, save_ext=None, **kwargs):
        super().__init__(env=env, **kwargs)
        # Checks
        assert save_ext or self.outfilepath, ("Either a save_ext or an "
                                              "outfilepath must be specified")

        self.infilepath_ext = save_ext or self.outfilepath.suffix
        self.context = context

    def build(self, complete=False):
        # Check that the string to write is in the parameters
        strings = [f for f in self.parameters if isinstance(f, str)]
        if len(strings) == 0:
            msg = ("The '{}' builder expects a single string as the "
                   "infilepath to save the temporary file.")
            raise BuildError(msg.format(self.__class__.__name__))
        string = strings[0]

        outfilepath = self.outfilepath
        logging.debug("Saving temporary file to '{}'".format(outfilepath))
        outfilepath.write_text(string)

        self.build_needed(reset=True)  # reset build flag
        return self.status



"""
URL routes for the Tornado server
"""
from tornado.web import url

from .handlers import (TreeHandler, CheckerHandler, SignalHandler,
                       CustomStaticFileHandler, PygmentizeHandler,
                       server_static_path)


url_patterns = [
    url(r"/", TreeHandler, name="tree"),
    url(r"/checkers", CheckerHandler, name="checkers"),
    url(r"/signals", SignalHandler, name="signals"),
    url(r"/media/(.*)", CustomStaticFileHandler, {'path': server_static_path}),
    url(r"/(.*\.dm)", PygmentizeHandler, name='disseminate_source'),
    url(r"/(.*\.tex)", PygmentizeHandler, name='latex_source',
        kwargs={'reload_projects': True}),
    url(r"/(.*\.html)", CustomStaticFileHandler,
        kwargs={'path': '.', 'reload_projects': True}),
    url(r"/(.*\.epub)", CustomStaticFileHandler,
        kwargs={'path': '.', 'reload_projects': True}),
    url(r"/(.*)", CustomStaticFileHandler, {'path': '.'}),
]



"""
A TargetBuilder for html files.
"""
from .target_builder import TargetBuilder
from ..copy import Copy
from ...paths import TargetPath
from ... import settings


class HtmlBuilder(TargetBuilder):
    """A builder for html files."""

    available = True
    priority = 1000
    infilepath_ext = '.dm'
    outfilepath_ext = '.html'
    ext_wildcards = {'*' + ext for ext in settings.tracked_deps['.html']}

    def __init__(self, env, context, **kwargs):
        super().__init__(env=env, context=context, **kwargs)

        # Add copy builders for extra dependencies, like css files
        target_root = env.target_root
        cache_path = env.cache_path
        target = self.outfilepath_ext

        for ext in self.ext_wildcards:
            copy_subbuilders = []
            for subbuilder in self.subbuilders:
                # Get the filepaths found from the render builder. These
                # should be copied over as needed. (ex: '.css' files)
                filepaths = [filepath for filepath in subbuilder.infilepaths
                             if filepath.match(ext)]

                for filepath in filepaths:
                    subpath = filepath.subpath

                    # Place the copied file either in the target_root directory
                    # or the cache_path directory, depending on whether
                    # self.use_cache is enabled (True).
                    root_path = cache_path if self.use_cache else target_root

                    # Reconstruct the path for the copied file in its
                    # destination directory
                    target_filepath = TargetPath(target_root=root_path,
                                                 target=target,
                                                 subpath=subpath)

                    # Create the copy builder and place it in the copy builders
                    # list to be added to the parallel builder
                    copy = Copy(env, parameters=filepath,
                                outfilepath=target_filepath)
                    copy_subbuilders.append(copy)

            # Add these copy builders to the parallel builders
            self._parallel_builder.subbuilders += copy_subbuilders



"""
Checkers for external dependencies.
"""

from .checker import Checker
from ..software_dependencies import image_deps


class ImageExtChecker(Checker):
    """Checker for external dependencies for image processing."""

    def __init__(self, image_deps=image_deps):
        super().__init__('image external deps', *image_deps.dependencies)



"""
Utilities for manipulating files and paths
"""
import os
import shutil
import logging


def parents(path):
    """Generate a list of parent paths for the given path.

    Parameters
    ----------
    path : str
        A filepath or directory.

    Returns
    -------
    parents : List[str]
        The parents of the path

    Examples
    --------
    >>> parents('/home/jlorieau/Documents/test.txt')
    ['/home/jlorieau/Documents', '/home/jlorieau', '/home']
    >>> parents('tex/chapter1/chapter1.tex')
    ['tex/chapter1', 'tex']
    """
    # Determine if path is a directory or a file
    is_dir = os.path.splitext(path)[1] == ""

    current_path = path if is_dir else os.path.split(path)[0]

    parents = []
    while current_path not in ('/', '.', ''):
        parents.append(current_path)
        current_path = os.path.split(current_path)[0]

    return parents


def link_or_copy(src, dst):
    """Create a hard link, if possible, between a src filepath to a dst
    filepath, or copy if a link is not possible.
    """
    logging.debug("Linking file '{}' -> '{}'".format(str(src), str(dst)))

    # Determine whether dst exists and whether is has the same inode as src
    src_inode = os.stat(src).st_ino
    try:
        dst_inode = os.stat(dst).st_ino
        dst_exists = True  # the destination file exists exist
    except FileNotFoundError:
        dst_inode = None
        dst_exists = False  # the destination file doesn't exist

    dst_different = (src_inode != dst_inode)

    # Remove the destination file if it exists and it's different from the
    # source
    if dst_exists and dst_different:
        os.remove(dst)
    elif dst_exists:
        # In this case, the destination exists and the destination is the same
        # file (by inode). The objective is already achieved and nothing else
        # needs to be done
        return None

    # Try create a hard link, if possible
    try:
        return os.link(src, dst)
    except OSError:
        return shutil.copyfile(src, dst)



"""
Utilities for lists.
"""
import hashlib
from itertools import filterfalse


def uniq(lst, key=None):
    """Find the unique items in the given list while preserving the order of
    the list and optionally using a key mapping function.

    Parameters
    ----------
    lst : list
        The list to check for duplicates.
    key : Optional[Callable]
        A mapping function that takes a list item and returns an alternate
        value to compare for unique entries.

    Examples
    --------
    >>> uniq([1, 2, 2, 3, 3, 3, 4, 5, 6, 6, 7])
    [1, 2, 3, 4, 5, 6, 7]
    >>> uniq([(1, 'a'), (2, 'a'), (3, 'b'), (4, 'c')], key=lambda i:i[1])
    [(1, 'a'), (3, 'b'), (4, 'c')]
    >>> uniq([(1, 'a'), (2, 'a'), (3, 'b'), (4, 'c')], key=lambda i:i[0])
    [(1, 'a'), (2, 'a'), (3, 'b'), (4, 'c')]
    """
    return list(unique_everseen(lst, key=key))


def unique_everseen(iterable, key=None):
    """List unique elements, preserving order. Remember all elements ever seen.
    """
    # unique_everseen('AAAABBBCCDAABBB') --> A B C D
    # unique_everseen('ABBCcAD', str.lower) --> A B C D
    seen = set()
    seen_add = seen.add
    if key is None:
        for element in filterfalse(seen.__contains__, iterable):
            seen_add(element)
            yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen_add(k)
                yield element


def dupes(lst, key=None):
    """Find duplicates in a list using, optionally, a key mapping function.

    Parameters
    ----------
    lst : list
        The list to check for duplicates.
    key : Optional[Callable]
        A mapping function that takes a list item and returns an alternate
        value to compare for duplicates.

    Examples
    --------
    >>> dupes([1, 2, 2, 3, 3, 3, 4, 5, 6, 6, 7])
    [2, 2, 3, 3, 3, 6, 6]
    >>> dupes([(1, 'a'), (2, 'a'), (3, 'b'), (4, 'c')], key=lambda i:i[1])
    [(1, 'a'), (2, 'a')]
    >>> dupes([(1, 'a'), (2, 'a'), (3, 'b'), (4, 'c')], key=lambda i:i[0])
    []
    """
    counts = dict()
    for item in lst:
        mapping = item if key is None else key(item)
        counts[mapping] = counts.setdefault(mapping, 0) + 1

    if key is None:
        return [item for item in lst if counts[item] > 1]
    else:
        return [item for item in lst if counts[key(item)] > 1]


def unwrap(lst):
    """Unwrap lists with only a single item.

    Parameters
    ----------
    lst : list
        The list to unwrap

    Returns
    -------
    unwrapped_list : list
        The unwrapped list

    Examples
    --------
    >>> unwrap([1, 2, 3])
    [1, 2, 3]
    >>> unwrap([1])
    1
    >>> unwrap([[1]])
    1
    >>> unwrap([[[3]]])
    3
    """
    new_l = lst[0] if isinstance(lst, list) and len(lst) == 1 else lst
    if isinstance(new_l, list) and len(new_l) == 1:
        new_l = unwrap(new_l)
    return new_l


def flatten(lst):
    """Flatten a list of lists.

    Examples
    --------
    >>> list(flatten([1, 2, 3]))
    [1, 2, 3]
    >>> list(flatten([[1, 2, 3], [4, 5, 6]]))
    [1, 2, 3, 4, 5, 6]
    >>> list(flatten([[1,2], [3, [4, 5], 6]]))
    [1, 2, 3, 4, 5, 6]
    """
    for element in lst:
        if isinstance(element, list):
            yield from flatten(element)
        else:
            yield element


def chunks(lst, N):
    """Return a generator in chunks of N.

    Parameters
    ----------
    lst : Union[List, Tuple]
        The iterable to break into chunks.
    N : int
        The size of chunks.

    Returns
    -------
    Generator
        A generator with sublists of items from 'l' broken into chunks of 'N'.
    """
    for i in range(0, len(lst), N):
        yield lst[i:i + N]


def md5hash(lst):
    """Generates a hash of a list or tuple.

    Examples
    --------
    >>> md5hash(('a', 'b', 1))
    '68b6a776378decbb4a79cda89087c4ce'
    >>> md5hash(('a', 'b', '1'))
    '68b6a776378decbb4a79cda89087c4ce'
    >>> md5hash(['a', [1, 'b'], 'c'])
    '1c81219649292a2ef9240fc997353078'
    >>> md5hash(['a', ['1', 'b'], 'c'])
    '1c81219649292a2ef9240fc997353078'
    """
    s = ''.join(md5hash(i) if isinstance(i, list) or isinstance(i, tuple)
                else str(i) for i in lst).encode()
    return hashlib.md5(s).hexdigest()



"""
The label base class for the label manager.
"""


class Label(object):
    """A label used for referencing.

    Parameters
    ----------
    doc_id : str
        The unique identifier (for documents within a label manager) for the
        document that owns the label.
    id : str
        The unique identifier(s) of the label. ex: 'nmr_introduction'. This
        should be unique for the entire project.
    kind : Tuple[str]
        The kind of the label is a tuple that identifies the kind of a label
        from least specific to most specific. ex: ('figure',), ('chapter',),
        ('equation',), ('heading', 'chapter',)
    order : Optional[Tuple[int]]
        The order/number of the label. The order is a tuple of integers with a
        length that matches the 'kind' tuple. Each entry represents the
        order/number of each kind  ex: for a kind ('heading', 'chapter') could
        have an order of (3, 2) which would represent the 3rd 'heading' and
        2nd 'chapter' item. Some of the orders may be reset, but the first item
        should not--e.g. the 'heading' count should represent the running count
        of all headings, while the chapter count may be reset. This ensures
        that the order of labels is preserved when the counter of sub-kinds are
        reset.
        (see :class:`OrderLabels
        <disseminate.label_manager.processors.OrderLabels>`)
    """

    doc_id = None
    id = None
    kind = None
    order = None

    def __init__(self, doc_id, id, kind, order=None):
        self.doc_id = doc_id
        self.id = id
        self.order = order

        # Wrap the kind in a tuple, if it's a string
        self.kind = (kind,) if isinstance(kind, str) else kind

    def __repr__(self, **params):
        cls_name = self.__class__.__name__
        if all(isinstance(i, tuple) or isinstance(i, list)
               for i in (self.kind, self.order)):
            kind_str = ", kind: "
            kind_str += ", ".join(['{}[{}]'.format(k, o)
                                   for k, o in zip(self.kind, self.order)])
        else:
            kind_str = ''
        msg = "{}(doc_id: '{}', id: '{}'{}".format(cls_name, self.doc_id,
                                                   self.id, kind_str)
        if params:
            attr_str = ", ".join(
                "{}: '{}'".format(k, v) for k, v in sorted(params.items()))
            msg += " " + attr_str + ")"
        else:
            msg += ")"
        return msg

    def __lt__(self, other):
        return (self.doc_id, self.id) < (other.doc_id, other.id)

    def __le__(self, other):
        return (self.doc_id, self.id) <= (other.doc_id, other.id)

    def __gt__(self, other):
        return (self.doc_id, self.id) > (other.doc_id, other.id)

    def __ge__(self, other):
        return (self.doc_id, self.id) >= (other.doc_id, other.id)

    def __eq__(self, other):
        return (self.doc_id, self.id) == (other.doc_id, other.id)

    def __ne__(self, other):
        return (self.doc_id, self.id) != (other.doc_id, other.id)

    def __hash__(self):
        return hash((self.__class__.__name__, self.doc_id, self.id))

    @property
    def number(self):
        """The number for the label's most specific kind."""
        if isinstance(self.order, tuple) and len(self.order) > 0:
            return self.order[-1]
        else:
            return None



"""
A request handler for rendering the project document tree.
"""
from datetime import datetime

from .server import ServerHandler
from .projects import load_projects


class TreeHandler(ServerHandler):
    """A request handler for the project document tree"""

    def get(self):
        # Load the project roo documents
        projects = load_projects(self.application)
        tree = self.tree_to_dict(projects)

        self.render("tree.html", projects=tree)

    def tree_to_dict(self, docs, level=1):
        """Convert the root documents into a list of dicts, suitable for
        rendering
        in the view.
        """
        doc_list = []

        for number, doc in enumerate(docs, 1):
            if level == 1:
                number = 1
            d = dict()

            # Set the number and level for the doc
            d['number'] = number
            d['level'] = level

            # Get metainformation and information on the source file
            d['title'] = doc.title
            d['src_filepath'] = doc.src_filepath
            d['project_root'] = doc.project_root

            # Get information on the targets
            targets = list(doc.targets.keys())  # ex: ['.html', '.pdf']
            target_links = [doc.target_filepath(target)
                            for target in targets]
            d['targets'] = {target: target_link for target, target_link in
                            zip(targets, target_links)}

            # Get information on the modification time for the source
            mtime = doc.src_filepath.stat().st_mtime
            d['date'] = datetime.fromtimestamp(mtime)

            # See if there are sub-documents to render as well
            # subdocuments is a dict with the src_filepath of the subdoc as a
            # key
            # and the subdoc itself as a vlue
            subdocs = doc.subdocuments

            if subdocs is not None:
                d['subdocs'] = self.tree_to_dict(subdocs.values(),
                                                 level=level + 1)

            doc_list.append(d)

        return doc_list



"""
The 'setup' CLI sub-command.
"""
import click

from .checkers import print_checkers
from .signals import print_signals
from ...signals.signals import signals


@click.command()
@click.option('--check', default=False, is_flag=True,
              help='Check required executables and packages')
@click.option('--list-signals', default=False, is_flag=True,
              help='List the available signals')
def setup(check, list_signals):
    """Setup and configuration options"""
    if check:
        print_checkers()
    elif list_signals:
        print_signals(signals)



"""
A TargetBuilder for pdf files.
"""
from .target_builder import TargetBuilder
from .tex_builder import TexBuilder


class PdfBuilder(TargetBuilder):
    """A builder for Pdf files."""

    available = True
    priority = 1000
    infilepath_ext = '.dm'
    outfilepath_ext = '.pdf'

    add_parallel_builder = False
    add_render_builder = False

    _tex_builder = None
    _pdf_builder = None

    def __init__(self, env, context, parameters=None, outfilepath=None,
                 subbuilders=None, **kwargs):
        # Setup the subbuilders
        subbuilders = subbuilders or []

        # Find the tex_builder or create one.
        builders = context.setdefault('builders', dict())
        if '.tex' not in builders:
            tex_builder = TexBuilder(env=env, context=context, target='tex',
                                     **kwargs)

            builders['.tex'] = tex_builder
        else:
            tex_builder = builders['.tex']

        self._tex_builder = tex_builder
        subbuilders.append(tex_builder)

        # Now add a Pdf converter
        tex2pdf_cls = self.find_builder_cls(in_ext='.tex', out_ext='.pdf')
        tex2pdf = tex2pdf_cls(env=env, target='pdf', use_media=False,
                              use_cache=True, **kwargs)
        self._pdf_builder = tex2pdf
        subbuilders.append(tex2pdf)

        # And a copy builder
        copy_builder_cls = self.find_builder_cls(in_ext='.*', out_ext='.*')
        copy_builder = copy_builder_cls(env=env, target='pdf', use_cache=False,
                                        **kwargs)
        subbuilders.append(copy_builder)

        super().__init__(env=env, context=context, parameters=parameters,
                         outfilepath=outfilepath, subbuilders=subbuilders,
                         **kwargs)

        # Setup the paths
        tex2pdf.parameters = [tex_builder.outfilepath]

        copy_builder.parameters = [tex2pdf.outfilepath]
        copy_builder.outfilepath = self.outfilepath

    def add_build(self, parameters, outfilepath=None, context=None, **kwargs):
        return self._tex_builder.add_build(parameters=parameters,
                                           outfilepath=outfilepath,
                                           context=context, **kwargs)



from .environment import Environment
from .builder import Builder
from .executor import executor
from . import target_builders
from . import (validators, exceptions, builder, pdf2svg, pdfcrop, scalesvg,
               copy, jinja_render, composite_builders, scanners, imagemagick,
               pdflatex, pdfrender, svgrender, latexmk, asy_builders,
               xhtml2epub)

__all__ = ('Environment', 'Builder', 'executor', 'target_builders',
           'validators', 'exceptions', 'builder', 'pdf2svg', 'pdfcrop',
           'scalesvg', 'copy', 'jinja_render', 'composite_builders',
           'scanners', 'imagemagick', 'pdflatex', 'pdfrender', 'svgrender',
           'latexmk', 'asy_builders', 'xhtml2epub')



"""
Generic types.
"""


class PositionalValue(object):
    """A placeholder for a positional value in a dict."""
    pass


class IntPositionalValue(PositionalValue):
    """A placeholder for a positional *integer* value in a dict."""
    pass


class FloatPositionalValue(PositionalValue):
    """A placeholder for a positional *float* value in a dict."""
    pass


class StringPositionalValue(PositionalValue):
    """A placeholder for a positional *string* value in a dict."""
    pass


def ispositional(p, positionalvalue_type=PositionalValue):
    """True, if the given parameter is a PositionalValue class or subclass.

    Parameters
    ----------
    p : object
        The parameter to test.
    positionalvalue_type : :class:`PositionalValue <.types.PositionalValue>`
        The PositionalValue class to test whether parameter is a subclass of
        this type.

    Returns
    -------
    bool
        True if parameter 'p' is the class or a subclass of PositonalValue
        (positionalvalue_type).

    Examples
    --------
    >>> ispositional('test')
    False
    >>> ispositional(IntPositionalValue)
    True
    >>> ispositional(PositionalValue)
    True
    """
    return isinstance(p, type) and issubclass(p, positionalvalue_type)


def positionalvalue_type(p):
    """Find the PostionalValue type for the given parameter.

    Parameters
    ----------
    p : object
        The parameter for which the PositionalValue class or subclass will be
        evaluated and returned.

    Returns
    -------
    positionalvalue_type : :class:`PositionalValue <.types.PositionalValue>`
        The PositionalValue class or subclass that matches the given parameter.

    Examples
    --------
    >>> positionalvalue_type(3)
    <class 'disseminate.utils.types.IntPositionalValue'>
    >>> positionalvalue_type('test')
    <class 'disseminate.utils.types.StringPositionalValue'>
    >>> positionalvalue_type('23')
    <class 'disseminate.utils.types.IntPositionalValue'>
    >>> positionalvalue_type('3.23')
    <class 'disseminate.utils.types.FloatPositionalValue'>
    >>> positionalvalue_type('src/media/image.png')
    <class 'disseminate.utils.types.StringPositionalValue'>
    """
    mappings = ((int, IntPositionalValue),
                (float, FloatPositionalValue),
                (str, StringPositionalValue),
                )

    for t, v in mappings:
        try:
            t(p)
            return v
        except ValueError:
            pass

    return PositionalValue



"""
The 'render' CLI sub-command.

.. note:: Implementing a progressbar using click or tqdm sometimes crashes
          the program without an exception and has crashed the terminal. I
          suspect it has to do with the redraw implementation. For this reason,
          these tools aren't used for the progress bar.
"""
import click

from .options import file_options, check_out_dir
from .utils.progressbar import ProgressTable
from ..builders.environment import Environment


@click.command()
@file_options
@click.option('-p', '--progress', is_flag=True, default=False,
              help="Show a progress bar for the build")
def build(in_path, out_dir=None, progress=False):
    """Build a disseminate project"""
    # Setup the build environment
    envs = Environment.create_environments(root_path=in_path,
                                           target_root=out_dir)
    docs = [env.root_document for env in envs]
    check_out_dir(root_docs=docs, out_dir=out_dir)

    for env in envs:
        root_builder = env.create_root_builder()
        builders = root_builder.flatten()

        if progress:  # Print progress, if enabled
            progress_table = ProgressTable(environment=env)
            progress_table.print_hdr()
            progress_table.print_row(builders)

        status = root_builder.status
        while status in {'ready', 'building'}:
            status = root_builder.build(complete=False)

            if progress:  # Print progress, if enabled
                progress_table.print_row(builders)
        print('Build:', root_builder.status)



from .label import Label
from ...utils.classes import weakattr


def get_label_number(label, default=''):
    """Get the number for the given label"""
    order = getattr(label, 'order', None)
    if isinstance(order, tuple) and len(order) > 0:
        return order[-1]
    else:
        return default


class ContentLabel(Label):
    """A label for content, like a heading (chapter, section, subsection) or
    an item that should show up in a table of contents, like a figure caption.

    Attributes
    ----------
    document_label : :obj:`DocumentLabel <.types.DocumentLabel>`
        The label object for the document that owns this label.
    chapter_label : Optional[:obj:`ContentLabel <.types.ContentLabel>`]
        The label for the chapter under which this label is under.
    section_label : Optional[:obj:`ContentLabel <.types.ContentLabel>`]
        The label for the section under which this label is under.
    subsection_label : Optional[:obj:`ContentLabel <.types.ContentLabel>`]
        The label for the subsection under which this label is under.
    subsubsection_label : Optional[:obj:`ContentLabel <.types.ContentLabel>`]
        The label for the subsubsection under which this label is under.
    """

    title = None

    document_label = weakattr()
    title_label = weakattr()
    part_label = weakattr()
    chapter_label = weakattr()
    section_label = weakattr()
    subsection_label = weakattr()
    subsubsection_label = weakattr()

    def __init__(self, doc_id, id, kind, title, order=None):
        super().__init__(doc_id=doc_id, id=id, kind=kind, order=order)
        self.title = title

    def __repr__(self):
        if self.title is not None:
            return super().__repr__(title=self.title)
        else:
            return super().__repr__()

    @property
    def part_number(self):
        """The number for the part to which this label belongs."""
        return get_label_number(self.part_label)

    @property
    def part_title(self):
        """The title for the part to which this label belongs."""
        return getattr(self.part_label, 'title', '')

    @property
    def chapter_number(self):
        """The number for the chapter to which this label belongs."""
        return get_label_number(self.chapter_label)

    @property
    def chapter_title(self):
        """The title for the chapter to which this label belongs."""
        return getattr(self.chapter_label, 'title', '')

    @property
    def section_number(self):
        """The number for the section to which this label belongs."""
        return get_label_number(self.section_label)

    @property
    def section_title(self):
        """The title for the section to which this label belongs."""
        return getattr(self.section_label, 'title', '')

    @property
    def subsection_number(self):
        """The number for the subsection to which this label belongs."""
        return get_label_number(self.subsection_label)

    @property
    def subsection_title(self):
        """The title for the subsection to which this label belongs."""
        return getattr(self.subsection_label, 'title', '')

    @property
    def subsubsection_number(self):
        """The number for the subsubsection to which this label belongs."""
        return get_label_number(self.subsubsection_label)

    @property
    def subsubsection_title(self):
        """The title for the subsubsection to which this label belongs."""
        return getattr(self.subsubsection_label, 'title', '')

    @property
    def tree_number(self):
        """The string for the number for the chapter, section, subsection and
        so on.

        e.g. Section 3.2.1."""
        # Get a tuple of the numbers, remove empty string items and None
        numbers = filter(bool, (self.part_number,
                                self.chapter_number,
                                self.section_number,
                                self.subsection_number,
                                self.subsubsection_number))

        # Convert the numbers to strings
        numbers = map(str, numbers)

        # Return a string with the numbers joined by a character.
        return '.'.join(numbers)



"""
A document corresponds to a single disseminate source file. A document may
contain sub-documents, which in turn may hold sub-documents, to form a tree.
The top-level document is known as a *root document* and the root document and
all sub-documents are known as a project.

Responsibilities
----------------

A document has the following responsibilities:

1. *Managers*. The root document is responsible to create managers.
   There should only be one of each type of manager in a project, and a
   project uses the same managers.
2. *Rendering*. The document is responsible for delegating rendering of the
   document in the target format(s).
3. *Processors*. When rendering a document, the 'body' entry in the context
   (*i.e.* the body of the disseminate source file) will be run through a
   list of processors, listed in the document class.
4. *Conversion*. Some documents types must be compiled or converted to their
   final format--*e.g.* to produce ``pdf`` targets, a ``tex`` target must be
   converted using ``pdflatex``. The document manages this conversion as part
   of the rendering process.
5. *Context*. A document owns a context that is populated by the disseminate
   file header, and which holds the variables used in rendering a document into
   a target format.
6. *Paths*. The document manages paths for the disseminate source file and the
   target files.

Definitions
-----------
1. *Root document*. A document that is not included as a sub-document to
   another document. The root document's context will have objects, like the
   label manager, that are shared by its sub-documents in a project.
2. *Project*. A root document and zero or more sub-documents forming a document
   tree. Sub-documents are included by ``include`` statements in the header of
   the root document.
3. *doc_id*. The document identifier for a document. It is unique for a
   document in a project.


"""

from .document import Document
from .document_context import DocumentContext
from . import exceptions, signals, receivers

__all__ = ('Document', 'DocumentContext', 'exceptions', 'signals', 'receivers')



"""
Receivers for processing a tag's contents on tag creation.
"""
import regex

from ..signals import tag_created
from ..exceptions import TagError
from ...utils.string import group_strings
from ... import settings

re_open_tag = regex.compile(settings.tag_prefix +  # tag character, '@'
                            r'(?P<tag>[A-Za-z0-9][\w]*)'
                            r'(?P<attributes>\[[^\]]+\])?'
                            r'(?P<open>{)?')
re_brace = regex.compile(r'[}{]')


@tag_created.connect_via(order=200)
def process_content(tag, tag_factory, **kwargs):
    """A receiver to parse the contents of tags into sub-tags."""
    content = parse_tags(content=tag.content, context=tag.context,
                         tag_factory=tag_factory)
    tag.content = content
    return tag


def parse_tags(content, context, tag_factory, level=1):
    """Process a string into a tree of tags, strings and lists of both.

    .. note:: The parsing of tags is idempotent. It will reprocess the
              generated tag without making changes.

    Parameters
    ----------
    content : Union[str, List[Union[str, :obj:`Tag <.Tag>`]]]
        A string to parse for tags, a list of strings and tags
    context : dict
        The context with values for the document.
    level : Optional[int]
        The current level of the ast.

    Returns
    -------
    tag_list : List[Union[str, :obj:`Tag <.Tag>`]]

    Raises
    ------
    TagError : :exc:`TagError <.exceptions.TagError>`
        Raises an TagError if the max depth has been achieved
        (settings.tag_max_depth). This is an attempt to foil the Billion Laughs
        attack.
    """
    # Conduct initial tests
    if level >= settings.tag_max_depth:
        msg = ("The maximum depth of '{}' has been reached in the Tag tree. "
               "Additional levels can be set by the 'settings.tag_max_depth'.")
        raise TagError(msg.format(settings.tag_max_depth))

    new_content = []

    def parse(x):
        return parse_tags(x, context, tag_factory, level + 1)

    # Look at the element and process it depending on whether it's a string,
    # tag or list
    if isinstance(content, str):
        pass
    elif isinstance(content, list):
        return list(map(parse, content))
    elif hasattr(content, 'content'):
        if isinstance(content.content, list):
            content.content = list(map(parse, content.content))
        else:
            content.content = parse(content.content)
        return content
    else:
        msg = "Tag content in an unknown format. The tag contents are: {}"
        raise TagError(msg.format(content))

    # The following only processes text
    text = content

    # The parser starts at the start of the text string
    position = 0

    # find open tags
    match_tag = re_open_tag.search(text[position:])

    # Process the tag
    while match_tag:
        # Add the text up to this tag the match start/end are relative to the
        # truncated text[position:] string, so the match start position has
        # to be offset when referencing the full 'text' string
        match_tag_start = position + match_tag.start()
        sofar = text[position:match_tag_start]  # the string up to this point
        if sofar:  # only add the sofar string if it isn't an empty string
            new_content.append(sofar)

        # Push up the position to the end of the tag match
        position += match_tag.end()
        start_position = position

        # Parse the tag contexts
        d = match_tag.groupdict()
        tag_name = d['tag']
        tag_attributes = d['attributes']

        # Find open and close braces and advance the position
        # up until the match closing brace is found
        brace_level = 1 if d['open'] is not None else 0
        match = re_brace.search(text[position:])
        while match and 0 < brace_level < 10:
            # Increment or decrement the match
            if match.group() == '}':
                brace_level -= 1
            elif match.group() == '{':
                brace_level += 1

            position += match.end()

            # Get the next match
            match = re_brace.search(text[position:])

        # Raise an error if the brace wasn't closed
        if brace_level > 0:
            msg = "The tag '{}' was not closed."
            raise TagError(msg.format(tag_name))

        # Parse the ast for the tag's content
        if d['open'] is None:
            # For tags with no open/close braces, then the content is empty
            tag_content = ''

        else:
            # Otherwise process the text within the tag's braces
            inner_text = text[start_position:position - 1]

            # Process the inner_text if it's more than the empty string
            tag_content = inner_text if inner_text else ''

        tag = tag_factory.tag(tag_name=tag_name,
                              tag_content=tag_content,
                              tag_attributes=tag_attributes,
                              context=context, )
        new_content.append(tag)

        # Find the next tag
        match_tag = re_open_tag.search(text[position:])

    # Add the remainer
    remainder = text[position:]
    if remainder:  # only add the remainder if it isn't an empty string
        new_content.append(remainder)

    # Remove empty strings
    group_strings(new_content)

    # Simplify the new_content list
    if len(new_content) == 1:
        # Unwrap new_content if it's a list with only one item
        new_content = new_content[0]
    elif len(new_content) == 0:
        # If the new_content list has no items, then just use an empty string
        new_content = ''

    return new_content



"""
Function(s) for setting the label order attributes.
"""

from collections import Counter


def reorder_and_purge_labels_by_doc_id(labels, root_context):
    """Reorder labels by doc_id and remove labels that refer to a doc_id for
    a document that no longer exists.

    Parameters
    ----------
    labels : Dict[Tuple[str,str], :obj:`Label <.label_manager.types.Label>`]
        A list of labels where the key is the (doc_id, label_id) and the values
        are the label objects.
    root_context : :obj:`.document.DocumentContext`
        The context for the root document.
    """

    # Get the order of doc_ids from the root_context
    root_doc = root_context.root_document
    if root_doc is not None:
        # Use the doc_ids from the project
        doc_ids = root_doc.doc_ids if root_doc is not None else []
    else:
        # Use the doc_ids from the labels themselves
        doc_ids = [label.doc_id for label in labels.values()]

    # Remove labels that aren't listed in the doc_ids
    filtered_labels = filter(lambda l: l.doc_id in doc_ids, labels.values())

    # Sort filtered labels by doc_id
    reordered_labels = sorted([(count, label)
                               for count, label in enumerate(filtered_labels)],
                              key=lambda k: (doc_ids.index(k[1].doc_id), k[0]))

    # Repopulate the labels dict (which should be an ordered dict)
    labels.clear()
    for count, label in reordered_labels:
        labels[label.doc_id, label.id] = label


def register_orders(labels, root_context, **kwargs):
    """Set the order attributes of labels.

    - The order is a tuple of integers that indicate the order or a label
      for the given kind.

      ex: a label may have a kind = ('heading', 'chapter') and an order
      of (3, 2). This indicates that the label is the 3rd heading and the
      2nd chapter.

    - This function may use the 'label_resets' entry in the context.
      This should be a dict with keys that are the kinds that reset other
      kind counters. The values are iterables of kinds whose counters are
      reset.

      ex: {'chapter': {'section', 'subsection', 'subsubsection'}}
      In this case, labels with a kind of 'chapter' will reset the counter
      for the 'section', 'subsection' and 'subsubsection' counters.

      Note that the 'heading' kind is not reset, and its order is maintained
      for the correct ordering of labels, even if the chapter/section/
      subsection counts are reset.

    - This function uses reorder_and_purge_labels_by_doc_id, which removes
      labels with a doc_id for documents that no longer exist.

    Parameters
    ----------
    labels : Dict[Tuple[str,str], :obj:`Label <.label_manager.types.Label>`]
        A list of labels where the key is the (doc_id, label_id) and the values
        are the label objects.
    root_context : :obj:`.document.DocumentContext`
        The context for the root document.
    """
    # Reorder labels
    reorder_and_purge_labels_by_doc_id(labels, root_context)

    # Keep track of the counts for each kind
    count = Counter()

    # Retrieve the label_resets.
    if 'label_resets' in root_context:
        reset_counts = root_context['label_resets']
    else:
        reset_counts = dict()

    # Process labels with a kind listed.
    filtered_labels = [label for label in labels.values()
                       if label.kind is not None]
    for label in filtered_labels:

        # Get the count for each of the kind items
        order = []
        for k in label.kind:
            # Increment the count for the kind
            count[k] += 1

            # Reset counters, if specified
            if k in reset_counts:
                for i in reset_counts[k]:
                    count[i] = 0

            # Append the count
            order.append(count[k])

        label.order = tuple(order)



"""
Signals for processing labels.
"""
from ..signals import signal

label_register = signal('label_register', doc="A signal sent when labels are "
                        "to be registered. Receivers take registered labels "
                        "(a list of labels) and collected labels (dict of "
                        "labels)")



"""
Tag event signals
"""
import weakref
from types import MethodType

from ..signals import signal

tag_created = signal('tag_created', doc=("A signal emitted when a tag is "
                                         "created. Receivers take a tag "
                                         "parameter."))


def emit(self, tag, **kwargs):
    """A custom emitter that checks which receivers to run based on attributes
    in the tag object."""
    return_values = []
    for order, receiver in sorted(self.receivers.items()):
        # De-reference receiver, if needed
        receiver = (receiver() if isinstance(receiver, weakref.ref) else
                    receiver)

        # Only run the receiver if it exists and the tag doesn't have
        # an attribute with the receiver's name as False.
        # ex: tag.process_content = False
        #     will not run the 'process_content' receiver
        if (receiver is not None and
           getattr(tag, receiver.__name__, True)):  # here is the difference
            return_values.append(receiver(tag=tag, **kwargs))
        else:
            return_values.append(None)
    return return_values


tag_created.emit = MethodType(emit, tag_created)



"""
Tags for lists.
"""
from math import ceil

import regex

from .tag import Tag
from ..attributes import Attributes
from ..formats import xhtml_list
from .. import settings


# Find lists with entries labeled with a '-' or '*' bullet for each list item
re_list = regex.compile(r'^\s*[\*\-]', regex.MULTILINE)

# Regex to clean list items with multi-line line-items
re_multiline_items = regex.compile(r'\s*\n+\s+')


def parse_string_list(s):
    r"""Parse a string with lists.

    Parameters
    ----------
    s : str
        The string to parse

    Returns
    -------
    parsed_list : List[Tuple[int, str]]

    Examples
    --------
    >>> parse_string_list("- This is my first item.\n"
    ...                   "  - This is my first subitem\n")
    [(0, 'This is my first item.'), (2, 'This is my first subitem')]
    """
    parsed_list = []
    current_match = None
    current_pos = 0
    list_level = 1

    for match in re_list.finditer(s):
        # Skip the first entry, since this is the string contents before
        # the start of the list
        if current_match is None:
            current_match = match
            current_pos = match.end()

            # Find the level. The number of spaces determines the level
            list_level = match.group().count(' ')
            continue

        # Retrieve the string up to this point. Remove leading spaces.
        end_pos = match.start()
        sub_string = s[current_pos:end_pos].strip()

        # Create the new entry in the returned list
        parsed_list.append((list_level, sub_string))

        # Set the current position and current match for the next iteration
        current_match = match
        current_pos = match.end()
        list_level = match.group().count(' ')

    # Add the remaining match to the returned list
    if current_match is not None:
        # Retrieve the string up to this point up until the end.
        # Remove leading spaces.
        end_pos = current_match.end()
        sub_string = s[end_pos:].strip()

        # Create the new entry in the returned list
        parsed_list.append((list_level, sub_string))

    return parsed_list


def clean_string_list(parsed_list):
    r"""Clean the string list created by parse_string_list.

    Cleaning include removeing extra spaces and newlines in parse string line
    elements.

    Parameters
    ----------
    parsed_list : List[Tuple[int, str]]
        The parsed list from parse_string_list.

    Returns
    -------
    cleaned_list : List[Tuple[int, str]]
        The cleaned list.

    >>> l = parse_string_list("- This is my first item.\n"
    ...                   "  - This is my first subitem\n")
    >>> clean_string_list(l)
    [(0, 'This is my first item.'), (2, 'This is my first subitem')]
    """
    return [(level, re_multiline_items.sub(" ", s))
            if isinstance(s, str) else
            (level, s)
            for level, s in parsed_list]


def normalize_levels(parsed_list,
                     list_level_spaces=settings.list_level_spaces):
    r"""Normalize the levels from a parse_list so that the first level is 0,
    and subsequent levels are properly incremented.

    Parameters
    ----------
    parsed_list : List[Tuple[int, str]]
        The parsed list from parse_string_list.
    list_level_spaces : int
        The number of spaces used to identify sub-levels in a list.

    Returns
    -------
    normalized_list : List[Tuple[int, str]]
        The normalized list with levels fixed.

    >>> l = parse_string_list("- This is my first item.\n"
    ...                       "  - This is my first subitem\n")
    >>> normalize_levels(l)
    [(0, 'This is my first item.'), (1, 'This is my first subitem')]
    """
    min_level = min(level for level, item in parsed_list) if parsed_list else 0
    return [(ceil((level - min_level) / list_level_spaces), item)
            for level, item in parsed_list]


def parse_list(content, context):
    """Parse lists (and sublists) from a string or list of content.

    Parameters
    ----------
    content : Union[str, List[Union[str, list, :obj:`Tag \
        <disseminate.tags.tag.Tag>`], :obj:`Tag <disseminate.tags.tag.Tag>`]
        The contents of the tag. It can either be a string, a tag or a list
        of strings, tags and lists.
    context : :obj:`.document.DocumentContext`
        The document's context dict.

    Returns
    -------
    returned_list : list
        The list of parse list items, to be used for the content of a list.
    """
    # Unify the format of the cotent
    if isinstance(content, str):
        content = [content]
    elif isinstance(content, Tag):
        content.content = parse_list(content.content)
    elif isinstance(content, list):
        pass
    else:
        return []

    # Go item-by-item to find list items
    returned_list = []
    current_list = None

    for item in content:
        # Add non-strings to a list
        if not isinstance(item, str) and current_list is not None:
            # If it's not a string, just add it to the current list
            current_list.append(item)

        # If it's a string, see if it has list items in it
        parsed_list = parse_string_list(item)
        parsed_list = clean_string_list(parsed_list)

        if parse_list:
            for level, list_item in parsed_list:
                if current_list is not None:
                    current_list.append(list_item)
                else:
                    current_list = [list_item]

                # Create the list item and add it the returned list
                content = (current_list[0] if len(current_list) == 1 else
                           current_list)
                li = ListItem(name='listitem', content=content,
                              attributes='level={}'.format(level),
                              context=context)
                returned_list.append(li)

                # Reset the current_list
                current_list = []

    return returned_list


class ListItem(Tag):
    """An item in a list"""

    active = False
    html_name = "li"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set the level to an integer
        self.attributes['level'] = int(self.attributes['level'])

    def tex_fmt(self, **kwargs):
        list_level = self.attributes['level']
        tex_content = super().tex_fmt(**kwargs)
        return "§" * (list_level + 1) + " " + tex_content + "\n"


class List(Tag):
    """A tag for lists"""

    active = True
    html_name = "ul"
    tex_env = "easylist"
    list_style = 'itemize'

    def __init__(self, name, content, attributes, context, **kwargs):
        self.name = name
        self.attributes = Attributes(attributes)
        self.context = context

        # Parse the content string into list items
        parsed_list = parse_string_list(content)
        parsed_list = clean_string_list(parsed_list)
        parsed_list = normalize_levels(parsed_list)

        # Create the list item tags
        self.content = [ListItem(name='listitem', content=list_content,
                                 attributes='level={}'.format(level),
                                 context=context)
                        for level, list_content in parsed_list]

    def tex_fmt(self, attributes=None, **kwargs):
        return super().tex_fmt(attributes=self.list_style, **kwargs)

    def html_fmt(self, content=None, attributes=None, format_func='html_fmt',
                 method='html', level=1, **kwargs):
        elements = []

        for tag in self.content:
            listlevel = tag.attributes['level']

            tag_html = getattr(tag, format_func)(method=method,
                                                 level=level + 1)
            elements.append((listlevel, tag_html))
        return xhtml_list(*elements, attributes=attributes,
                          listtype=self.html_name, method=method, level=level)


class OrderedList(List):
    """A tag for ordered lists"""

    active = True
    html_name = "ol"
    aliases = ('outline',)
    list_style = 'enumerate'



"""
Image tags
"""
import pathlib

from .tag import Tag, TagError
from .utils import xhtml_percentwidth, tex_percentwidth
from ..signals import signal
from ..paths.utils import find_files
from ..formats import tex_cmd

add_file = signal('add_file')


class ImgFileNotFound(TagError):
    """The image file was not found."""
    pass


class Img(Tag):
    """The img tag for inserting images.

    When rendering to a target document format, this tag may use a converter,
    the attributes and context of this tag to convert the infile to an outfile
    in a needed format.

    Attributes
    ----------
    active : bool
        If True, the Tag can be used by the TagFactory.
    html_name : str
        If specified, use this name when rendering the tag to html. Otherwise,
        use name.
    in_ext : Optional[str]
        Used to correctly identify the builder to add if the contents need
        to be rendered first.
    img_filepath : str
        The path for the (source) image.
    """

    active = True

    process_content = False
    process_typography = False

    html_name = 'img'
    in_ext = None
    _infilepath = None
    _outfilepaths = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._outfilepaths = dict()

    def content_as_filepath(self, content=None, context=None):
        """Returns a filepath from the content, if it's a valid filepath,
        or returns None if it isn't.

        Parameters
        ----------
        content : Optional[Union[str, List[Union[str, list, :obj:`Tag \
            <.tag.Tag>`], :obj:`Tag <.tag.Tag>`]]
            The contents of the tag. It can either be a string, a tag or a list
            of strings, tags and lists.
        context : Optional[:obj:`Type[BaseContext] <.BaseContext>`]
            The context with values for the document.

        Returns
        -------
        filepath : Union[:obj:`pathlib.Path`. None]
            The filepath, if found, or None if a valid filepath was not found.
        """
        content = content or self.content
        context = context or self.context

        # Move the contents to the infilepath attribute
        if isinstance(content, pathlib.Path) and content.is_file():
            return content
        elif isinstance(content, str):
            # Get the infilepath for the file
            filepaths = find_files(content, context)
            return filepaths[0] if filepaths else None

        return None

    def add_file(self, target, content=None, attributes=None, context=None):
        """Convert and add the file dependency for the specified document
        target.

        Parameters
        ----------
        target : str
            The document target format. ex: '.html', '.tex'
        content : Optional[Union[str, List[Union[str, list, :obj:`Tag \
            <.tag.Tag>`], :obj:`Tag <.tag.Tag>`]]
            The contents of the tag. It can either be a string, a tag or a list
            of strings, tags and lists.
        attributes : Optional[Union[str, \
            :obj:`Attributes <.attributes.Attributes>`]]
            The attributes of the tag.
        context : Optional[:obj:`Type[BaseContext] <.BaseContext>`]
            The context with values for the document.

        Returns
        -------
        outfilepath : :obj:`.paths.TargetPath`
            The filepath for the file in the target document directory.

        Raises
        ------
        ImgFileNotFound
            Raises an ImgFileNotFound exception if a filepath couldn't be
            found in the tag contents.
        BuildError
            If a builder could not be found for the builder
        """
        # See if a cached path exists already. This can only be done if the
        # content, attributes and context aren't specified because the
        # values of this tag will be used for these parameters.
        if all(i is None for i in (content, attributes, context)):
            if target in self._outfilepaths:
                return self._outfilepaths[target]
            else:
                can_cache = True
        else:
            can_cache = False

        # Retrieve the unspecified arguments
        content = content or self.content
        attrs = attributes or self.attributes
        context = context or self.context

        # Prepare the parameters. Either their a filepath of the contents
        # or the contents themselves.
        content = (self.content_as_filepath(content=content,
                                            context=context) or
                   content)
        parameters = [content] + list(attrs.filter(target=target).totuple())

        # Use the content's filepath suffix as the in_ext, if a file has
        # been
        # specified, otherwise use this class's in_ext attribute
        in_ext = (content.suffix if isinstance(content, pathlib.Path) else
                  self.in_ext)
        outfilepaths = add_file.emit(parameters=parameters, context=context,
                                     in_ext=in_ext, target=target,
                                     use_cache=False)
        assert len(outfilepaths) == 1

        outfilepath = outfilepaths[0]

        # Cache the outfilepath, if possible
        if can_cache:
            self._outfilepaths[target] = outfilepath

        return outfilepath

    def tex_fmt(self, content=None, attributes=None, context=None, **kwargs):
        # Add the file dependency
        outfilepath = self.add_file(target='.tex', content=content,
                                    context=context, attributes=attributes)

        # Format the width
        attrs = attributes or self.attributes.copy()
        attrs = tex_percentwidth(attrs, target='.tex')

        # Get the filename for the file.
        base = outfilepath.with_suffix('')
        suffix = outfilepath.suffix

        # If the filename has unicode characters, you need to detokenize them
        # for latex to run correctly
        try:
            str(outfilepath).encode('ascii')
        except UnicodeEncodeError:
            base = tex_cmd(cmd='detokenize', formatted_content=str(base))

        # Wrap this filename in curly braces, which is needed for paths with
        # spaces and absolute paths
        dest_filepath = "{{{base}}}{suffix}".format(base=base, suffix=suffix)

        return tex_cmd(cmd='includegraphics', attributes=attrs,
                       formatted_content=str(dest_filepath))

    def html_fmt(self, content=None, attributes=None, context=None,
                 method='html', **kwargs):
        # Add the file dependency
        target = '.' + method if not method.startswith('.') else method
        outfilepath = self.add_file(target=target, content=content,
                                    context=context, attributes=attributes)
        url = outfilepath.get_url(context=self.context, target=target)

        # Format the width and attributes
        attrs = attributes or self.attributes.copy()
        attrs = xhtml_percentwidth(attrs, target=target)
        attrs['src'] = url

        return super().html_fmt(content='', attributes=attrs, method=method,
                                **kwargs)



"""
Navigation tags
"""
import weakref
from os.path import relpath
from itertools import groupby

from markupsafe import escape

from .tag import Tag
from .ref import Ref
from ..signals import signal


document_tree_updated = signal('document_tree_updated')


@document_tree_updated.connect_via(order=10000)
def set_navigation_labels(root_document):
    """Set the navigation labels in the context of all documents in a document
    tree."""
    root_context = root_document.context
    assert root_context.is_valid('label_manager')

    # Get doc_ids by target
    doc_ids_by_target = dict()
    documents = root_document.documents_list(only_subdocuments=False,
                                             recursive=True)
    for document in documents:
        targets = document.targets
        doc_id = document.doc_id

        for target in targets:
            lst = doc_ids_by_target.setdefault(target, [])
            lst.append(doc_id)

    # Get all of the document labels
    label_manager = root_context.get('label_manager')
    labels = label_manager.get_labels_by_kind(kinds='heading')
    labels_by_doc_id = {doc_id: next(group) for doc_id, group in
                        groupby(labels, lambda l: l.doc_id)}

    # Set the prev_labels and next_labels entries in the context of each
    # document
    for target in targets:
        if target not in doc_ids_by_target:
            continue
        doc_ids = doc_ids_by_target[target]

        for document in documents:
            doc_id = document.doc_id

            # Find this document's doc_id in the doc_ids_by_target
            try:
                position = doc_ids.index(doc_id)
            except ValueError:
                position = None

            # See if the prev, next, curr links are avalaiable
            for name, rel in (('prev', - 1), ('curr', 0), ('next', 1)):

                # Don't use back-tracking indexes
                rel_position = position + rel if position is not None else None
                rel_position = (None if rel_position is None or
                                rel_position < 0 else rel_position)
                key = "_".join((name, target.strip('.')))
                try:
                    other_doc_id = doc_ids[rel_position]
                    other_label = labels_by_doc_id[other_doc_id]
                    document.context[key] = weakref.ref(other_label)
                except (IndexError, KeyError, TypeError):
                    # If the other document could not be found, make sure this
                    # entry isn't in the context. This can happen if the
                    # context contains stale information from a previous
                    # invocation of this function.

                    if key in document.context:
                        del document.context[key]


class Next(Ref):
    """Produces links to the next document.

    Notes
    -----
    1. As opposed to the parent Ref tag, if a label is not found, an empty
       string is returned by the target format functions.
    2. The following are tag attributes that are recognized:

    kind : str
        The kind of label to reference by this tag. By default, it's a
        heading.
    """

    active = True

    def __init__(self, name, content, **kwargs):
        # Bypass the Ref initiator, as it sets the label_id
        # Do not use the tag contents--this tag cannot do anything with it.
        content = ''
        Tag.__init__(self, name=name, content=content, **kwargs)

    def default_fmt(self, **kwargs):
        # The next reference is designed for html only
        return ""

    def tex_fmt(self, cache=None, **kwargs):
        context = self.context
        cache = dict() if cache is None else cache

        if 'label' not in cache and 'next_tex' in context:
            label = context['next_tex']
            cache['label'] = label()  # Dereference. See above.

        if 'label' in cache:
            return super(Next, self).html_fmt(cache=cache, **kwargs)
        else:
            return ""

    def html_fmt(self, cache=None, **kwargs):
        context = self.context
        cache = dict() if cache is None else cache

        if 'label' not in cache and 'next_html' in context:
            label = context['next_html']
            cache['label'] = label()  # Dereference. See above.

        if 'label' in cache:
            return super(Next, self).html_fmt(cache=cache, **kwargs)
        else:
            return ""

    def xhtml_fmt(self, **kwargs):
        raise NotImplementedError


class Prev(Ref):
    """Produces links to the previous document.

    Notes
    -----
    1. As opposed to the parent Ref tag, if a label is not found, an empty
       string is returned by the target format functions.
    2. The following are tag attributes that are recognized

    kind : str
        The kind of label to reference by this tag. By default, it's a
        heading.
    """

    active = True

    def __init__(self, name, content, **kwargs):
        # Bypass the Ref initiator, as it sets the label_id
        # Do not use the tag contents--this tag cannot do anything with it.
        content = ''
        Tag.__init__(self, name=name, content=content, **kwargs)

    def default_fmt(self, **kwargs):
        # The next reference is designed for html only
        return ""

    def tex_fmt(self, cache=None, **kwargs):
        context = self.context
        cache = dict() if cache is None else cache

        if 'label' not in cache and 'prev_tex' in context:
            label = context['prev_tex']
            cache['label'] = label()  # Dereference. See above.

        if 'label' in cache:
            return super(Prev, self).html_fmt(cache=cache, **kwargs)
        else:
            return ""

    def html_fmt(self, cache=None, **kwargs):
        context = self.context
        cache = dict() if cache is None else cache

        if 'label' not in cache and 'prev_html' in context:
            label = context['prev_html']
            cache['label'] = label()  # Dereference. See above.

        if 'label' in cache:
            return super(Prev, self).html_fmt(cache=cache, **kwargs)
        else:
            return ""

    def xhtml_fmt(self, **kwargs):
        raise NotImplementedError


class OtherLink(Tag):
    """Produces a link to a source document or another target for this
    document."""

    active = False
    other_target = None

    def __init__(self, name, content, **kwargs):
        # Bypass the Ref initiator, as it sets the label_id
        # Do not use the tag contents--this tag cannot do anything with it.
        content = ''
        Tag.__init__(self, name, content, **kwargs)

    def this_filepath(self, target):
        """The filepath for this (target) document."""
        context = self.context

        if (context is not None and target is not None and
           target in context.targets):
            return context.target_filepath(target)
        else:
            return None

    def other_filepath(self, target=None):
        """The filepath for the other document target or src_filepath to create
        a link to."""
        context = self.context
        target = target or self.other_target

        if (context is not None and target is not None and
           target in context.targets):
            return context.target_filepath(target)
        else:
            return None

    def other_relpath(self, this_target, other_target=None):
        """Produce the the relpath for the other target."""
        # Get the filepaths
        this_filepath = self.this_filepath(target=this_target)
        other_filepath = self.other_filepath(target=other_target)

        if this_filepath is None or other_filepath is None:
            # Both paths should be found to forumate a relative path
            return ''

        if this_filepath == other_filepath:
            # If both paths are the same, the relpath is the current dir,
            # and this should not give a link to itself.
            return ''

        # Generate the relative path
        rp = relpath(other_filepath, this_filepath.parent)
        return escape(rp)

    def default_fmt(self, **kwargs):
        return ""

    def tex_fmt(self, **kwargs):
        return self.other_relpath(this_target='.tex')

    def html_fmt(self, **kwargs):
        return self.other_relpath(this_target='.html')

    def xhtml_fmt(self, **kwargs):
        raise NotImplementedError


class Srclink(OtherLink):
    """Produces a link to the src file for this document"""
    active = True

    def other_filepath(self, target=None):
        context = self.context
        return (context['src_filepath'] if context is not None and
                'src_filepath' in context else None)


class Txtlink(OtherLink):
    """Produces a link to the txt target for this document."""
    active = True
    other_target = '.txt'


class Texlink(OtherLink):
    """Produces a link to the tex target for this document."""
    active = True
    other_target = '.tex'


class Pdflink(OtherLink):
    """Produces a link to the pdf target for this document"""
    active = True
    other_target = '.pdf'


class Epublink(OtherLink):
    """Produces a link to the epub target for the root document."""
    active = True
    other_target = '.epub'

    def other_filepath(self, target=None):
        context = self.context
        root_doc = getattr(context, 'root_document', None)
        root_context = root_doc.context if root_doc is not None else None

        if root_context is not None and '.epub' in root_context.targets:
            return root_context.target_filepath('.epub')
        else:
            return None



"""
Receivers for processing a tag's typography.
"""
import regex

from ..signals import tag_created


@tag_created.connect_via(order=300)
def process_typography(tag, tag_base_cls, **kwargs):
    """A receiver to parse the typography of tags."""
    return process_tag_typography(tag=tag, tag_base_cls=tag_base_cls)


re_endash = regex.compile(r"[\s\u00a0]*(--|\u2013)[\s\u00a0]*")
re_emdash = regex.compile(r"[\s\u00a0]*(---|\u2014)[\s\u00a0]*")
re_apostrophe = regex.compile(r"(?<=\w)'(?=\w)")
re_single_start = regex.compile(r"(?<!\w)'(?=\S)")
re_single_end = regex.compile(r"(?<=\S)'(?!\w)")
re_double_start = regex.compile(r"(?<!\w)\"(?=\S)")
re_double_end = regex.compile(r"(?<=\S)\"(?!\w)")


def process_tag_typography(tag, tag_base_cls, level=1):
    """Process the typography for an AST.

    .. note:: This function should be run after process_ast.

    Parameters
    ----------
    tag : Union[str, list]
        A string to parse into an AST, a list of strings or an existing AST.
    tag_base_cls : :class:`Tag <.Tag>`
        The base class for Tag objects.
    level : Optional[int]
        The current level of the tag.

    Returns
    -------
    tag : :obj:`Tag <.Tag>`
        The AST is a root tag with a content comprising a list of tags or
        strings.
    """
    if isinstance(tag, str):
        # Process the tag if it's simply a string
        tag = re_emdash.sub('\u2014', tag)
        tag = re_endash.sub('\u2013', tag)
        tag = re_apostrophe.sub('’', tag)
        tag = re_single_start.sub('‘', tag)
        tag = re_single_end.sub('’', tag)
        tag = re_double_start.sub('“', tag)
        tag = re_double_end.sub('”', tag)
        return tag

    elif (isinstance(tag, tag_base_cls) and
          getattr(tag, 'process_typography', False)):
        # Process the tag contents
        tag.content = process_typography(tag=tag.content,
                                         tag_base_cls=tag_base_cls,
                                         level=level + 1)
        return tag

    elif isinstance(tag, list):
        # Process each element of lists
        for i, element in enumerate(tag):
            tag[i] = process_typography(tag=element,
                                        tag_base_cls=tag_base_cls,
                                        level=level + 1)

    return tag



"""
Wrappers and utilities for formatting text and tags into different formats
"""

from .tex import tex_env, tex_cmd, tex_verb, TexFormatError
from .xhtml import xhtml_tag, xhtml_entity, xhtml_list, XHtmlFormatError

__all__ = ('tex_env', 'tex_cmd', 'tex_verb', 'TexFormatError',
           'xhtml_tag', 'xhtml_entity', 'xhtml_list', 'XHtmlFormatError',)



"""
Command-line interface tools for Checkers
"""
from click import style

from ..term import fill_string
from ...checkers import Checker, All, Any, Optional

spaces_per_level = 2
width = 80


def print_single_check(msg, status, color=None, bold=False, spacer=' ',
                       level=1):
    """Print a single check line."""
    msg = (' ' * spaces_per_level * level) + msg  # indent
    status_str = '[{:^8}]  '.format(status)
    status_len = len(status_str)

    if color is not None:
        status_str = style(status_str, fg=color, bold=bold)
    return fill_string(msg, end=status_str, end_len=status_len, spacer=spacer,
                       width=width)


def print_dependencies(item, required=True, level=1):
    # Setup the status string
    available = getattr(item, 'available', None)
    if available is True:
        status_str, color, bold, spacer = 'PASS', 'green', True, ' '
    elif available is False and required:
        status_str, color, bold, spacer = 'FAIL', 'red', True, '.'
    elif available is False and not required:
        status_str, color, bold, spacer = 'MISSING', 'yellow', True, ' '
    else:
        status_str, color, bold, spacer = 'UNKNOWN', 'white', False, ' '

    if isinstance(item, All):
        # All are required
        msg = "Checking required dependencies for '{}'".format(item.category)
        msg = print_single_check(msg=msg, status=status_str, color=color,
                                 bold=bold, level=level, spacer=spacer)
        print(msg)
        for i in item:
            print_dependencies(item=i, required=True, level=level + 1)
    elif isinstance(item, Any):
        # One is required are required
        msg = ("Checking alternative dependencies for "
               "'{}'".format(item.category))
        msg = print_single_check(msg=msg, status=status_str, color=color,
                                 bold=bold, level=level, spacer=spacer)
        print(msg)
        for i in item:
            print_dependencies(item=i, required=False, level=level + 1)
    elif isinstance(item, Optional):
        # Optional
        # One is required are required
        msg = ("Checking alternative dependencies for "
               "'{}'".format(item.category))
        msg = print_single_check(msg=msg, status=status_str, color=color,
                                 bold=bold, level=level, spacer=spacer)
        print(msg)
        for i in item:
            print_dependencies(item=i, required=False, level=level + 1)
    elif hasattr(item, 'name'):
        msg = "Checking dependency '{}'".format(item.name)
        msg = print_single_check(msg=msg, status=status_str, color=color,
                                 bold=bold, level=level, spacer=spacer)
        print(msg)


def print_checkers():
    """Print the availabilitiy of executables and packages from the checkers.
    """
    # Get the checker subclasses
    checker_subclses = Checker.checker_subclasses()

    # Iterate through the checkers
    for checker_subcls in checker_subclses:
        checker = checker_subcls()

        # Do an overall check
        checker.check()

        # Print the checker status
        print_dependencies(checker)



from .composite_builder import CompositeBuilder
from ...paths.utils import find_file


class ParallelBuilder(CompositeBuilder):
    """A composite builder that runs subbuilders in parallell (i.e. run the
    subbuilders together at the same time)"""

    action = 'parallel build'
    parallel = True

    def build_needed(self, reset=False):
        return any(sb.build_needed(reset=reset) for sb in self.subbuilders)

    def add_build(self, parameters, outfilepath=None, target=None,
                  context=None, in_ext=None, out_ext=None, builder_cls=None,
                  **kwargs):
        """
        Add a subbuilder to the ParallelBuilder.

        Parameters
        ----------
        parameters, args : Tuple[:obj:`pathlib.Path`, str, tuple, list]
            The input parameters (dependencies), including filepaths, for the
            build
        outfilepath : Optional[str, :obj:`pathlib.Path`]
            The path for the output file.
        target : Optional[str]
            The document target for the build. ex: 'html' or 'tex'
        context : Optional[:obj:`.context.BaseContext`]
            The document context dict that owns the parallel builder.
        in_ext : Optional[str]
            The extension for the input file. This is used to find the correct
            builder with the find_builder_cls method. If this is not specified,
            an filepath in the parameters should be specified.
        out_ext : Optional[str]
            The extension for the input file. This is used to find the correct
            builder with the find_builder_cls method. If this is not specified,
            the outfilepath should be specified.
        builder_cls : Optional[:obj:`.builders.Builder`]
            The builder class to use in create a new build.
        **kwargs
            Optional keyword arguments to use in creating the builder.

        Returns
        -------
        builder : :obj:`.builders.Builder`
            The newly created builder.
        """
        # Setup the arguments
        context = context or self.env.context

        parameters = parameters or []
        parameters = (parameters if isinstance(parameters, list) or
                      isinstance(parameters, tuple) else [parameters])

        # Find the files in the parameters
        parameters = [find_file(i, context, raise_error=False) or i
                      for i in parameters]

        target = target or self.target

        # Find the builder class
        if builder_cls is None:
            # Get the input extensions for the builder in_ext
            if in_ext is None:
                in_exts = [infilepath for infilepath in parameters
                           if isinstance(infilepath, str) and
                           infilepath.startswith('.')]
                in_exts += [infilepath.suffix for infilepath in parameters
                            if hasattr(infilepath, 'suffix')]

                in_ext = in_exts[0] if in_exts else None

            # Get the output extension for the builder out_ext
            if out_ext is None:
                if (isinstance(outfilepath, str) and
                   outfilepath.startswith('.')):
                    out_ext = outfilepath
                elif hasattr(outfilepath, 'suffix'):
                    out_ext = outfilepath.suffix
                else:
                    out_ext = None

            builder_cls = self.find_builder_cls(in_ext=in_ext, out_ext=out_ext,
                                                target=target)

        # Create the builder
        builder = builder_cls(env=self.env, parameters=parameters,
                              outfilepath=outfilepath, context=context,
                              target=target, **kwargs)

        self.subbuilders.append(builder)

        return builder



"""
The manager for labels.
"""
from collections import OrderedDict
from threading import Lock

from .types import ContentLabel, DocumentLabel
from .exceptions import LabelNotFound, DuplicateLabel
from .register_orders import register_orders
from .register_content_labels import register_content_labels
from ..utils.classes import weakattr
from ..utils.dict import find_entry
from ..utils.string import replace_macros
from .. import settings


def parse_id(id, context=None):
    """Parses a label's identifier into a doc_id and label_id.

    Parameters
    ----------
    id : str
        The unique identifier (within a project of documents) for the label.
        The id includes the label_id, and it might include the doc_id.
        ex:
        'test.dm::intro'. (See label_sep for details)
        'ch:nmr-introduction'
    context : :obj:`.DocumentContext`
        The document context.

    Returns
    -------
    doc_id, label_id : Tuple[Union[str, None], str]
        The doc_id and label_id from the label id.

    Examples
    --------
    >>> parse_id('intro')
    (None, 'intro')
    >>> parse_id('fig:diagram-1')
    (None, 'fig:diagram-1')
    >>> parse_id('test1.dm::fig:diagram-1')
    ('test1.dm', 'fig:diagram-1')
    """
    if context is not None and 'label_sep' in context:
        label_sep = context['label_sep']
    else:
        label_sep = settings.default_context['label_sep']

    parts = id.split(label_sep)
    return tuple(parts) if len(parts) == 2 else (None, parts[0])


class LabelManager(object):
    """Manage labels for a project (a document tree).

    Parameters
    ----------
    root_context : :obj:`DocumentContext <.DocumentContext>`
        The context for the root document. The label manager does not own the
        context object, so only a weak reference to the context is stored.

    Attributes
    ----------
    labels : Dict[Tuple[str,str], :obj:`Label <.label_manager.types.Label>`]
        A list of labels where the key is the (doc_id, label_id) and the values
        are the label objects.
    """

    root_context = weakattr()
    labels = None
    collected_labels = None
    registered = False

    def __init__(self, root_context):
        self.labels = OrderedDict()
        self.root_context = root_context

    def register(self, context=None):
        """Register the labels.
        """
        if self.registered:
            return None

        # Only 1 thread should register the labels at a time
        lock = Lock()
        lock.acquire(blocking=True, timeout=10)
        context = context or self.root_context

        # Run the registration functions.
        for func in (register_orders,  # Register label 'order' attribute
                     register_content_labels,  # Set chapter/section attributes
                     ):
            func(labels=self.labels, root_context=context)

        # Labels have been registered. Release the lock
        self.registered = True
        if lock.locked():
            lock.release()

    def reset(self, doc_ids=None):
        """Reset the the labels.

        Parameters
        ----------
        doc_ids : Union[str, List[str], Tuple[str]]
            If specified, remove the labels for the given doc_ids
        """
        # Prepare the parameters
        doc_ids = {doc_ids} if isinstance(doc_ids, str) else doc_ids

        if doc_ids is None:
            self.labels.clear()
        else:
            keys_to_remove = set(filter(lambda k: k[0] in doc_ids,
                                        self.labels.keys()))
            for key in keys_to_remove:
                del self.labels[key]

        self.registered = False

    def add_label(self, id, kind, context, label_cls, *args, **kwargs):
        """Add a label.

        Parameters
        ----------
        id : str
            The unique identifier (within a project of documents) for the
            label. The id includes the label_id and possibly the doc_id of
            the label. 'test.dm::intro'. (See parse_id for details)
            ex: 'ch:nmr-introduction'
        kind : Union[str, List[str], Tuple[str]]
            The kind of the label. ex: 'figure', ('heading', 'chapter'),
            'equation'
        context : :obj:`DocumentContext <.DocumentContext>`
            The context for the document adding the label. (This may be
            different from the context of the root document, self.root_context)
        label_cls : :class:`Type[Label] <.label_manager.types.Label>`
            The label class (or subclass) to use in creating the label.

        Raises
        ------
        DuplicateLabel : :exc:`DuplicateLabel <.DuplicateLabel>`
            Raised if a label with the same id already exists in this
            label manager.

        """
        self.registered = False

        # Parse the label_id
        doc_id, label_id = parse_id(id, context=context)
        if doc_id is None:
            assert context.is_valid('doc_id')
            doc_id = context['doc_id']

        # Make the key for the labels dict
        label_key = (doc_id, label_id)

        # See if it's a duplicate
        if label_key in self.labels:
            other_label = self.labels[label_key]
            msg = "Label id '{}' already exists in the labels as '{}'."
            raise DuplicateLabel(msg.format(id, other_label))

        # Organize the kind into a tuple, if needed
        if isinstance(kind, str):
            kind = (kind,)

        # Now create the label and add it to the labels
        label = label_cls(doc_id=doc_id, id=label_id, kind=kind, order=None,
                          *args, **kwargs)
        self.labels[label_key] = label

        return label

    def add_content_label(self, id, kind, title, context):
        """Add content label.

        See :meth:`add_label` for usage details.
        """
        label = self.add_label(id=id, kind=kind, context=context,
                               label_cls=ContentLabel, title=title)
        label.title = title
        return label

    def add_document_label(self, id, kind, title, context):
        """Add document label.

        See :meth:`add_label` for usage details.
        """
        label = self.add_label(id=id, kind=kind, context=context,
                               label_cls=DocumentLabel, title=title)
        label.title = title
        return label

    def get_label(self, id, register=True, context=None):
        """Return the label for the given label id.

        .. note:: This function registers the added labels.

        Parameters
        ----------
        id : str
            The unique identifier (within a project of documents) for the
            label. The id includes the label_id and possibly the doc_id of
            the label. 'test.dm::intro'. (See parse_id for details)
            ex: 'ch:nmr-introduction'
        register : Optional[bool]
            If True, labels will be registered before doing the search
        context : :obj:`DocumentContext <.DocumentContext>`
            The context for the document adding the label. (This may be
            different from the context of the root document, self.root_context)

        Returns
        -------
        label : :obj:`Type[Label] <.label_manager.types.Label>`
            The corresponding label.

        Raises
        ------
        LabelNotFound : :exc:`LabelNotFound <.exceptions.LabelNotFound>`
            A LabelNotFound exception is raised if a label with the given id
            could not be found.
        """
        # Register the labels
        self.register()

        # Parse the label_id
        doc_id, label_id = parse_id(id, context=context or self.root_context)

        # Try the label key
        if (doc_id, label_id) in self.labels:
            return self.labels[(doc_id, label_id)]

        # Try to find the first label with a matching label_id, if no
        # doc_id is specified
        if doc_id is None:
            try:
                return next(l for k, l in self.labels.items()
                            if k[1] == label_id)
            except StopIteration:
                pass

        # I give up! I can't find the label.
        msg = "Could not find a label with identifier '{}'"
        raise LabelNotFound(msg.format(id))

    def get_labels_by_id(self, ids, register=True, context=None):
        """Return the labels with the given (optional) doc_ids and label_ids.

        Parameters
        ----------
        ids : Union[str, Tuple[str]]
            The unique identifiers (within a project of documents) for the
            label. The id includes the label_id and possibly the doc_id of the
            label. 'test.dm::intro'. (See parse_id for details)
            ex: 'ch:nmr-introduction'
        register : Optional[bool]
            If True, labels will be registered before doing the search
        context : :obj:`DocumentContext <.DocumentContext>`
            The context for the document adding the label. (This may be
            different from the context of the root document, self.root_context)

        Returns
        -------
        labels : List[:obj:`Type[Label] <.label_manager.types.Label>`]
            A list of label objects.

        Raises
        ------
        LabelNotFound : :exc:`LabelNotFound <.exceptions.LabelNotFound>`
            A LabelNotFound exception is raised if a label with the given id
            could not be found.
        """
        # Prepare the parameters
        ids = (ids if any(isinstance(ids, t) for t in (list, tuple, set))
               else [ids])

        # Register the labels
        self.register()

        return list(map(lambda x:
                        self.get_label(x, register=False, context=context),
                    ids))

    def get_labels_by_kind(self, doc_id=None, kinds=None, register=True):
        """Return a filtered and sorted list of all labels for the given
        document and kinds.

        .. note:: This function registers the added labels.

        Parameters
        ----------
        doc_id : Optional[str]
            If specified, only label for the given document id will be
            returned. (This is used an alternative to the context.)
        kinds : Optional[Union[str, List[str], Tuple[str]]
            If None, all label kinds are returned.
            If string, all labels matching the kind string will be returned.
            If a list of strings is returned, all labels matching all the kinds
            listed will be returned.
        register : Optional[bool]
            If True, labels will be registered before doing the search

        Returns
        -------
        labels : List[:obj:`Type[Label] <.label_manager.types.Label>`]
            A list of label objects.
        """
        # Prepare the parameters
        kinds = kinds or []
        kinds = (kinds if isinstance(kinds, list) or isinstance(kinds, tuple)
                 else [kinds])

        # Register the labels
        self.register()

        # Filter labels by doc_id.
        labels = self.labels.values()
        if doc_id is not None:
            labels = [label for label in labels if label.doc_id == doc_id]

        # Filter labels by kind
        returned_labels = []
        if kinds:
            for kind in kinds:
                returned_labels += [label for label in labels
                                    if kind in label.kind]
        else:
            returned_labels = labels

        return returned_labels

    def format_string(self, id, *keys, target=None):
        """Retrieve the formatted label string for a label.

        Label format strings are intended to be replaced with the values of the
        label by  replace_macros (:func:`replace_macros
        <disseminate.macros.replace_macros>`).

        Parameters
        ----------
        id : str
            The label of the label ex: 'ch:nmr-introduction'
        keys : Optional[List[str], Tuple[str]]
            If specified, use the given keys to find entries in the format_str
            dict. (see :func:`find_entry <disseminate.utils.dict.find_entry>`)
        target : Optional[str]
            If specified, try finding format strings for the given target.
        """
        label = self.get_label(id=id)

        dicts = []
        if 'label_fmts' in self.root_context:
            dicts.append(self.root_context['label_fmts'])

        # Construct the parameters for find_entry
        target = (target[1:] if isinstance(target, str) and
                  target.startswith('.') else target)

        # Find the label format string
        if len(keys) > 0:
            fmt_string = find_entry(dicts, *keys, suffix=target)
        else:
            fmt_string = find_entry(dicts, *label.kind, suffix=target)

        return replace_macros(fmt_string, {'@label': label})



"""
A builder to convert from TEX to PDF
"""
from .builder import Builder


class Pdflatex(Builder):
    """Compile a latex document into pdf using pdflatex"""

    action = ("pdflatex "
              "-interaction=nonstopmode "  # Do not hang on error
              "-halt-on-error "  # Do not hang on error
              "-output-directory={builder.cache_path} "  # dir for temp files
              "-jobname={builder.jobname} "  # filename of output file
              "{builder.infilepaths}")  # tex file to use

    available = True
    priority = 1000
    required_execs = ('pdflatex',)

    infilepath_ext = '.tex'
    outfilepath_ext = '.pdf'

    @property
    def cache_path(self):
        return self.outfilepath.parent

    @property
    def jobname(self):
        return self.outfilepath.stem



"""
CLI utilities
"""


def progress_builders(builders):
    """Returns the progress from a flattened list of builders.

    Parameters
    ----------
    builders : List[:obj:`.builders.Builder`]
        A flattened list of builders

    Returns
    -------
    done, total : Tuple[int, int]
        The number of done builders and the total number of builders.
    """
    return len([b for b in builders if b.status == 'done']), len(builders)


class SimpleTable(object):
    """A simple table to print to terminal.

    Parameters
    ----------
    *col_widths : Tuple[int]
        The widths of columns
    """

    col_widths = None

    def __init__(self, *col_widths):
        assert all(isinstance(width, int) for width in col_widths)
        self.col_widths = col_widths

    @property
    def width(self):
        """The number of columns for the table"""
        return sum(self.col_widths)

    def print_hdr(self, *values):
        """Print the table header."""
        assert len(self.col_widths) == len(values), ("The number of values "
                                                     "should match the number "
                                                     "of columns")
        hdr = ""
        for value, width in zip(values, self.col_widths):
            hdr += str(value).ljust(width, " ")
        print(hdr)
        bar = ""
        for value, width in zip(values, self.col_widths):
            value = str(value)
            bar += ("-" * len(value)).ljust(width, " ")
        print(bar)

    def print_row(self, *values):
        """Print a table row."""
        row = ""
        for value, width in zip(values, self.col_widths):
            row += value.ljust(width, " ")
        print(row)


class ProgressTable(SimpleTable):
    """A table for displaying the progress of a build.

    Parameters
    ----------
    environment : :obj:`.builders.Environment`
        The build environment

    Attributes
    ----------
    env_name : str
        The unique name of the build environment
    num_docs : int
        The number of documents in a build tree
    hdrs : Tuple[str]
        The names of columns
    spacing : int
        The number of spaces between columns.
    """

    env_name = None
    num_docs = None
    hdrs = ('Environment', 'Number Docs', 'Done / Total Builds (%)')
    spacing = 2

    def __init__(self, environment):
        self.env_name = environment.name
        root_doc = environment.root_document
        if root_doc is not None:
            docs_list = root_doc.documents_list(only_subdocuments=False,
                                                recursive=True)
            self.num_docs = str(len(docs_list))
        else:
            self.num_docs = ''

        # Get the column width
        col1_width = max(len(self.hdrs[0]), len(self.env_name)) + self.spacing
        col2_width = max(len(self.hdrs[1]), len(self.num_docs)) + self.spacing
        col3_width = len(self.hdrs[2]) + self.spacing

        super().__init__(col1_width, col2_width, col3_width)

    def print_hdr(self):
        super().print_hdr(*self.hdrs)

    def print_row(self, builders):
        """Print a row based on the status of the given builders."""
        done, total = progress_builders(builders)
        percent = (done / total) * 100.
        status_str = "{} / {} ({:.1f}%)".format(done, total, percent)
        rows = [self.env_name, self.num_docs, status_str]
        super().print_row(*rows)



"""
A receiver to process tag entries in a context
"""
from ..signals import document_onload
from ...tags import TagFactory


@document_onload.connect_via(order=10000)
def process_tags(context, **kwargs):
    """Convert context entries into tags for entries listed the
    process_context_tags' context entry.

    This function converts tags for context entries identified by keys listed
    in the 'process_context_tags' context entry

    .. note:: This function is designed to work with string macro entries.
              These are identified by context entries with keys that start
              with the settings.tag_prefix (e.g. '@test'). These *should not*
              be converted into asts, as they are required for simple string
              replacement.
    """
    assert context.is_valid('process_context_tags')

    # Find the tags to process, based on the context entries in the
    # 'process_context_tags' entry of the context and the available context
    # entries
    process_tags = context['process_context_tags']
    keys = set(context.keys())
    keys = keys.intersection(process_tags)

    for k in keys:
        value = context[k]

        # Only process strings
        if not isinstance(value, str):
            continue

        context[k] = TagFactory.tag(tag_name=k, tag_content=value,
                                    tag_attributes='', context=context)

    return context



from .server import ServerHandler, server_template_path, server_static_path
from .tree import TreeHandler
from .checkers import CheckerHandler
from .signals import SignalHandler
from .pygmentize import PygmentizeHandler
from .static import CustomStaticFileHandler

__all__ = ('ServerHandler', 'server_template_path', 'server_static_path',
           'TreeHandler', 'CheckerHandler', 'SignalHandler',
           'PygmentizeHandler', 'CustomStaticFileHandler')



"""Disseminate is a document processor and generation engine.
"""

from setuptools import setup, find_packages
from codecs import open
import os


here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get the version number and package information.  The __version__.py
# file is executed so that the disseminate package is not loaded.  At
# this point, the C/C++ extensions may not be built, and loading the
# package will lead to an ImportError. This approach circumvents this problem.
__version__ = None  # This is a version string
VERSION = None  # This is a 5-item version tuple
exec(open("./src/disseminate/__version__.py").read())

# Organize classifiers
if VERSION[3] == 'alpha':
    classifiers = ['Development Status :: 3 - Alpha', ]
elif VERSION[3] == 'beta':
    classifiers = ['Development Status :: 4 - Beta', ]
elif VERSION[3] == 'rc':
    classifiers = ['Development Status :: 4 - Beta', ]
elif VERSION[3] == 'final':
    classifiers = ['Development Status :: 5 - Production/Stable', ]
else:
    classifiers = []


classifiers += [
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: Science/Research',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Topic :: Text Processing',
    'Topic :: Text Processing :: General',
    'Topic :: Text Processing :: Markup',
    'Topic :: Text Processing :: Markup :: HTML',
    'Topic :: Text Processing :: Markup :: LaTeX']


setup(
    name='disseminate',  # Required
    version=__version__,
    description='A document processor and generation engine',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Justin L Lorieau',
    classifiers=classifiers,
    keywords='document processor academic publishing',
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,  # include MANIFEST.in
    install_requires=[
        'regex>=2018.11.22',         # No license, replaced with re
        'jinja2>=3.0',               # 3-clause BSD
        'lxml>=4.3.0',               # BSD license
        'python-slugify>=2.0.1',     # MIT license
        'pdfCropMargins>=0.1.4',     # GPL v3
        # The following are needed for the CLI
        'click>=7.0',                # 3-clause BSD license
        # The following are needed for the preview function
        'tornado>=6.1',
        # The following is needed for the @code tag
        'pygments>=2.6',             # BSD 2 license
        'numpy',
        'pandas',                     # 3-clause BSD
        'diskcache>=4.1',
        'pathvalidate>=2.2'],
    extras_require={  # Optional
        'dev': ['sphinx', 'sphinx_rtd_theme', 'sphinx-click', 'numpydoc',
                'asv'],
        'test': ['pytest', 'pytest-cov', 'tox', 'coverage', 'flake8',
                 'epubcheck>=0.4'],
        'termcolor': ['termcolor']  # MIT license
    },
    scripts=['scripts/dm', ],
    entry_points={
        'console_scripts': [
            'dm = disseminate.cli:main'],
        'pygments.lexers': [
            'dmlexer = disseminate.utils.pygments.dm:DmLexer'
        ]
    }
)



"""
A Builder to copy or link files.
"""
import logging
import pathlib
from shutil import copyfile, SameFileError

from .builder import Builder


class Copy(Builder):
    """A builder to copy or build a file."""

    action = 'copy'
    available = True
    active_requirements = ('priority',)
    priority = 1000
    infilepath_ext = '.*'
    outfilepath_ext = '.*'

    def __init__(self, env, parameters=None, outfilepath=None, **kwargs):
        # reset outfilepath_ext so that '.*' isn't used in automatically
        # generating the outfilepath (generate_outfilepath)
        self.outfilepath_ext = None

        super().__init__(env=env, parameters=parameters,
                         outfilepath=outfilepath, **kwargs)

        # Make sure the parameters are pathlib.Path objects
        filtered_parameters = []
        for parameter in self.parameters:
            if isinstance(parameter, str):
                parameter = pathlib.Path(parameter)

            if isinstance(parameter, pathlib.Path):
                filtered_parameters.append(parameter)
            else:
                filtered_parameters.append(parameter)
        self.parameters = filtered_parameters

    @property
    def status(self):
        if (not self.build_needed() or
           self.infilepaths and self.infilepaths[0] == self.outfilepath):
            # Done if the build is not needed or the src and dest are the
            # same file
            return "done"
        elif not all(i.exists() for i in self.infilepaths):
            # Missing if one or more of the parameters are missing
            return "missing (parameters)"
        else:
            return "ready"

    def build(self, complete=False):
        if self.status == 'ready':
            infilepath = self.parameters[0]
            outfilepath = self.outfilepath

            # Copy the file if the 2 files have different paths and contents.
            # If they have the same path, or they are hardlinks of each other,
            # then a shutil.SameFileError exception is raised.
            try:
                copyfile(infilepath, outfilepath)
                logging.debug("Copying '{}' -> '{}'".format(infilepath,
                                                            outfilepath))
            except SameFileError:
                pass

            self.build_needed(reset=True)  # reset build flag
        return self.status



"""
Builders that use ImageMagick's convert.
"""
from .builder import Builder


class ImageMagick(Builder):
    """An abstract base class for ImageMagick's convert."""

    action = 'convert {builder.infilepaths} {builder.outfilepath}'
    available = False
    priority = 10000
    required_execs = ('convert',)


class Tif2png(ImageMagick):
    """Converter for a tif file to a png file."""

    available = True
    infilepath_ext = '.tif'
    outfilepath_ext = '.png'


class Tiff2png(ImageMagick):
    """Converter for a tiff file to a png file."""

    available = True
    infilepath_ext = '.tiff'
    outfilepath_ext = '.png'



"""
Misc utilities for tags.
"""
from math import ceil
from copy import copy

from ..attributes import Attributes
from ..utils.types import StringPositionalValue


def content_to_str(content, target='.txt', **kwargs):
    """Convert a tag or string to a string for the specified target.

    This function is used to convert an element, which is either a string, tag
    or list of strings and tags, to a string.

    Parameters
    ----------
    content : Union[str, List[Union[str, list, :obj:`Tag \
        <disseminate.tags.tag.Tag>`], :obj:`Tag <disseminate.tags.tag.Tag>`]
        The element to convert into a string.
    target : str, optional
        The target format of the string to return.

    Returns
    -------
    formatted_str : str
        A string in the specified target format
    """
    target = target[1:] if target.startswith('.') else target  # rm leading '.'
    format_func = ('default_fmt' if target == 'txt' else
                   '_'.join((target, 'fmt')))  # ex: tex_fmt

    return ''.join(format_content(content=content, format_func=format_func,
                                  **kwargs))


def format_content(content, format_func, **kwargs):
    """Format the content using the format_func.

    Parameters
    ----------
    content : Union[str, List[Union[str, list, :obj:`Tag \
            <disseminate.tags.Tag>`]], optional
            The content to format
    format_func : str
        The name of the format_func to use. ex: 'tex_fmt'

    Returns
    -------
    content : str or list
        The content formatted using the specified format_func.
    """
    # Wrap content in a list and increment level
    content = [content] if not isinstance(content, list) else content

    content = [getattr(i, format_func)(**kwargs)
               if hasattr(i, format_func) else i for i in content]
    return content[0] if len(content) == 1 else content


def repl_tags(element, tag_class, replacement):
    """Replace all instances of a tag class with a replacement string.

    Parameters
    ----------
    element : Union[str, list :obj:`Tag <disseminate.tags.core.Tag>`]
        The element to replace tags with a replacement string.
    tag_class : :class:`Tag <disseminate.tags.core.Tag>`
        A tag class or subclass to replace.
    replacement : str
        The string to replace the tag with.
    """
    if isinstance(element, tag_class):
        return replacement
    elif hasattr(element, 'content'):
        element.content = repl_tags(element=element.content,
                                    tag_class=tag_class,
                                    replacement=replacement)
        return element
    elif isinstance(element, list):
        for i in range(len(element)):
            element[i] = repl_tags(element=element[i], tag_class=tag_class,
                                   replacement=replacement)

    return element


def replace_context(tag, new_context):
    """Replace the context for the given tag (and all subtags) to the given
    new_context.

    Parameters
    ----------
    tag : :obj:`Tag <disseminate.tags.core.Tag>`
        The tag to replace the context.
    new_context : :obj:`Type[BaseContext] <.context.BaseContext>`
        The new context to replace.
    """
    # Get all of the tags for the given tag
    flattened_tags = tag.flatten(filter_tags=True)

    # Replace the context for all tags
    for tag in flattened_tags:
        tag.context = new_context


def copy_tag(tag):
    """Create a copy of the given tag.

    The tag, attributes and content are deep copies, and the context points to
    the same context as the given tag.

    Parameters
    ----------
    tag : Union[str, list :obj:`Tag <disseminate.tags.core.Tag>`]
        The tag to copy.

    Returns
    -------
    tag_copy : :obj:`Tag <disseminate.tags.core.Tag>`
        The tag copy.
    """
    if isinstance(tag, str):
        # Do nothing for strings
        return tag
    if isinstance(tag, list):
        # Process each item in lists
        return [copy_tag(i) for i in tag]

    # At this point, the tag parameter should actually be a tag. First, copy
    # this tag
    tag_copy = copy(tag)  # shallow copy

    # make copies of the following fields. The __weakrefattrs__ dict needs
    # to be copied so that entries stored in it (like the context weak ref)
    # can be changed independently for the source and copied tags.
    for field in ('attributes', '__weakrefattrs__'):
        if hasattr(tag, field):
            value = getattr(tag, field)
            setattr(tag_copy, field, value.copy())

    # Copy its content
    tag_copy.content = copy_tag(tag_copy.content)

    return tag_copy


def percentage(value):
    """Given a strings (or floats or ints), convert into a float number between
    [0.0, 100.0].

    Parameters
    ----------
    value : Union[str, int, float]

    Returns
    -------
    percentage : Union[int, None]
        The percentage from 0 to 100, or
        None if the value could not be converted.

    Examples
    --------
    >>> percentage('10.2%')
    11
    >>> percentage('10.2')
    1020
    >>> percentage('0.3%')
    1
    >>> percentage('0.3')
    30
    >>> percentage(30)
    30
    """
    if isinstance(value, str):
        if '%' in value:
            try:
                value = float(value.strip().strip('%'))
            except ValueError:
                value = None
        else:
            try:
                value = float(value.strip()) * 100.0
            except ValueError:
                value = None

    try:
        return abs(ceil(value))
    except (ValueError, TypeError):
        return None


def tex_percentwidth(attributes, target='.tex', use_positional=False):
    """Generates an tex width string based on the 'width' entry in the
    attributes.

    Currently, this function works in halves (w50), thirds (w33, w66) and
    quarters (w25, w75)

    Parameters
    ----------
    attributes : :obj:`.attributes.Attributes`
        The attributes dict
    target : Optional[str]
        The target format to format the attribute dict for.
    use_positional : Optional[bool]
        Insert the entry as a positional attribute

    Returns
    -------
    attributes: :obj:`.attributes.Attributes`
        The same attributes dict with the width inserted as a
        StringPositionalArgument, if a width was found.

    Examples
    --------
    >>> tex_percentwidth('width=50%')
    Attributes{'width': '0.49\\\\textwidth'}
    >>> tex_percentwidth('width=3.2in')
    Attributes{'width': '3.2in'}
    >>> tex_percentwidth('width=50%', use_positional=True)
    Attributes{'width': '50%', \
'0.49\\\\textwidth': <class '...StringPositionalValue'>}
    """  # noqa: D301
    attributes = (Attributes(attributes)
                  if not isinstance(attributes, Attributes) else attributes)

    # Convert the width to a number between 0 and 100
    width = attributes.get('width', target=target)
    if width is None:
        return attributes

    percent_width = percentage(width)  # width from 0 to 100

    # Find which width window this falls in
    windows = {99: (75, 100),  # ]75, 100]
               74: (67, 75),
               65: (50, 67),
               49: (34, 50),
               32: (25, 33),
               24: (0, 25)}

    fit_percent_width = None
    if isinstance(percent_width, int):
        for test_percent_width, (range_min, range_max) in windows.items():
            if range_min < percent_width <= range_max:
                fit_percent_width = test_percent_width

    # Set the width value in the attributes dict
    if isinstance(fit_percent_width, int):
        str = "{}\\textwidth".format(fit_percent_width / 100.)
        if use_positional:
            attributes[str] = StringPositionalValue
        else:
            attributes['width'] = str
    else:
        str = "{}".format(width)
        if use_positional:
            attributes[str] = StringPositionalValue
        else:
            attributes['width'] = str

    return attributes


def xhtml_percentwidth(attributes, target='.html'):
    """Generates an (x)html class name based on the 'width' entry in the
    attributes.

    Currently, this function works in halves (w50), thirds (w33, w66) and
    quarters (w25, w75)

    Parameters
    ----------
    attributes : :obj:`.attributes.Attributes`
        The attributes dict
    target : Optional[str]
        The target format to format the attribute dict for.

    Returns
    -------
    attributes: :obj:`.attributes.Attributes`
        The same attributes dict with the 'class' entry set to an html class
        for the percent width, if a width was found.

    Examples
    --------
    >>> xhtml_percentwidth('width=50%')
    Attributes{'width': '50%', 'class': 'w50'}
    >>> xhtml_percentwidth('width=0%')  # None returned
    Attributes{'width': '0%'}
    """
    attributes = (Attributes(attributes)
                  if not isinstance(attributes, Attributes) else attributes)

    width = attributes.get('width', target=target)
    if width is None:
        return attributes

    percent_width = percentage(width)  # width from 0 to 100

    if not isinstance(percent_width, int):
        return attributes

    windows = {'w100': (75, 100),  # ]75, 100]
               'w75': (67, 75),
               'w66': (50, 67),
               'w50': (34, 50),
               'w33': (25, 33),
               'w25': (0, 25)}

    for html_class, (range_min, range_max) in windows.items():
        if range_min < percent_width <= range_max:
            cls_str = (attributes['class'] + ' ' + html_class
                       if 'class' in attributes else html_class)
            attributes['class'] = cls_str
            return attributes

    return attributes



"""
A generic request handler for server functions.
"""
import platform
from pathlib import Path
from traceback import extract_tb, format_list
from sys import executable as python_executable

from tornado.web import RequestHandler

from .projects import load_projects
from ...__version__ import __version__


# Get the path for the templates directory and add it to the bottle
# TEMPLATE_PATH
server_template_path = (Path(__file__).parent.parent.parent / 'templates' /
                        'server')
server_static_path = server_template_path / 'media'


class ServerHandler(RequestHandler):
    """A request handler for server functions"""

    def initialize(self, reload_projects=False, *args, **kwargs):
        if reload_projects:
            load_projects(self.application)
        super().initialize(*args, **kwargs)

    def set_default_headers(self):
        # Disable mimetype sniffing/guessing to allow viewing of plain/text
        self.set_header('X-Content-Type-Options', 'nosniff')

    def get_template_path(self):
        """The template path for the server templates"""
        return str(server_template_path)

    def write_error(self, status_code, **kwargs):
        """Render custom error pages"""
        # Add information on the current version of the software
        kwargs['version'] = __version__
        kwargs['python_version'] = platform.python_version()
        kwargs['python_executable'] = python_executable

        # Add information on the platform
        kwargs['platform'] = platform.platform()

        # Parse the exception and populate the message kwarg
        exc_info = kwargs.get('exc_info', None)

        if exc_info is not None and len(exc_info) == 3:
            # Unpack exc_info into the exception class, exception instance and
            # traceback
            exc_type, exc, tb = exc_info

            # Parse the exception
            kwargs['exc'] = exc
            kwargs['exc_type'] = exc_type.__name__
            kwargs['exc_filename'] = getattr(exc, 'filename', '')
            kwargs['exc_lineno'] = getattr(exc, 'lineno', '')
            kwargs['exc_msg'] = exc.args[0] if len(exc.args) > 0 else ''
            kwargs['exc_args'] = exc.args
            # Parse the traceback
            kwargs['traceback'] = format_list(extract_tb(tb))

        if status_code == 404:
            self.render("404.html", status_code=status_code, **kwargs)
        elif status_code == 500:
            self.render("500.html", status_code=status_code, **kwargs)
        else:
            self.render("error.html", status_code=status_code, **kwargs)



"""
Classes for different kinds of paths.
"""
# TODO: Implement equivalence tests.

import pathlib
import os
import re
import weakref


re_dbl_slash = re.compile(r'(?<!\:)//')  # match '//' but not '://'


# We need to derive the WindowsPath and PosixPath so that we can add
# entries into its __dict__.
class WindowsPath(pathlib.WindowsPath):
    pass


class PosixPath(pathlib.PosixPath):
    pass


class SourcePath(object):
    """A path for a file in the source directory that keeps track of the
    project_root and subpath.

    Parameters
    ----------
    project_root : Optional[str]
        The path for the project's root directory.
    subpath : Optional[str]
        The relative path within the project's root directory.
    """
    project_root = None
    subpath = None
    __mixclass__ = None

    def __new__(cls, project_root='', subpath=''):
        # A variety of objects, including strings, pathlib.Path object or other
        # objects can be passed to this method. It's simplest just to convert
        # them to a string before using them.
        project_root = str(project_root)
        subpath = str(subpath)

        subpath = pathlib.Path(subpath)
        assert not subpath.is_absolute(), ("The subpath argument cannot be an "
                                           "absolute path")

        # Create the mixclass, if needed, this is the class with either
        # PosixPath or WindowsPath mixed in with SourcePath
        if cls.__mixclass__ is None:
            flavor_cls = WindowsPath if os.name == 'nt' else PosixPath

            class MixClass(flavor_cls, SourcePath):
                pass
            MixClass.__name__ = cls.__name__
            cls.__mixclass__ = MixClass

        obj = pathlib.Path.__new__(cls.__mixclass__, project_root, subpath)
        obj.project_root = cls.__mixclass__(project_root, '')
        obj.subpath = cls.__mixclass__('', subpath)
        return obj

    def __copy__(self):
        return SourcePath(project_root=self.project_root, subpath=self.subpath)

    def __deepcopy__(self, memo):
        return SourcePath(project_root=self.project_root, subpath=self.subpath)

    def use_name(self, name):
        """with_name(name) that returns a SourcePath.

        A new method name is used since the flavor_cls method takes precedence.

        Examples
        --------
        >>> p = SourcePath(project_root='/media', subpath='tests/fig1.png')
        >>> p.use_name('fig2.png')
        SourcePath('/media/tests/fig2.png')
        >>> p.use_name('fig2.png').subpath
        SourcePath('tests/fig2.png')
        """
        return SourcePath(project_root=self.project_root,
                          subpath=self.subpath.with_name(name))

    def use_suffix(self, suffix):
        """with_suffix(suffix) that returns a SourcePath.

        A new method name is used since the flavor_cls method takes
        precedence.

        Examples
        --------
        >>> p = SourcePath(project_root='/media', subpath='tests/fig1.png')
        >>> p.use_suffix('.pdf')
        SourcePath('/media/tests/fig1.pdf')
        >>> p.use_name('fig1.pdf').subpath
        SourcePath('tests/fig1.pdf')
        """
        return SourcePath(project_root=self.project_root,
                          subpath=self.subpath.with_suffix(suffix))

    def use_subpath(self, subpath):
        """Returns a SourcePath with the subpath replaced with the given string
        or path.

        Examples
        --------
        >>> p = SourcePath(project_root='/media', subpath='tests/fig1.png')
        >>> p.use_subpath('text.txt')
        SourcePath('/media/text.txt')
        >>> p.use_subpath('text.txt').subpath
        SourcePath('text.txt')
        """
        return SourcePath(project_root=self.project_root,
                          subpath=subpath)


class TargetPath(object):
    """A path for a file in a target directory that keeps track of the
    target_root, target and subpath.

    Parameters
    ----------
    target_root : Optional[str]
        The path for the root directory for rendered files.
    target : Optional[str]
        The sub-directory for the target type to render. ex: 'html' or 'tex'
    subpath : Optional[str]
        The relative path within the target sub-directory.
    """
    target_root = None
    target = None
    subpath = None
    __mixclass__ = None

    def __new__(cls, target_root='', target='', subpath=''):
        # A variety of objects, including strings, pathlib.Path object or other
        # objects can be passed to this method. It's simplest just to convert
        # them to a string before using them.
        target_root = str(target_root)
        target = str(target)
        subpath = str(subpath)

        subpath = pathlib.Path(subpath)
        assert not subpath.is_absolute(), ("The subpath argument cannot be an "
                                           "absolute path")

        # Create the mixclass, if needed, this is the class with either
        # PosixPath or WindowsPath mixed in with TargetPath
        if cls.__mixclass__ is None:
            flavor_cls = WindowsPath if os.name == 'nt' else PosixPath

            class MixClass(flavor_cls, TargetPath):
                pass
            MixClass.__name__ = cls.__name__
            cls.__mixclass__ = MixClass

        if isinstance(target, str):
            target = target.strip('.')  # '.html' -> 'html'

        obj = pathlib.Path.__new__(cls.__mixclass__, target_root, target,
                                   subpath)
        obj.target_root = cls.__mixclass__(target_root, '', '')
        obj.target = cls.__mixclass__('', target, '')
        obj.subpath = cls.__mixclass__('', '', subpath)
        return obj

    def __copy__(self):
        return TargetPath(target_root=self.target_root, target=self.target,
                          subpath=self.subpath)

    def __deepcopy__(self, memo):
        return TargetPath(target_root=self.target_root, target=self.target,
                          subpath=self.subpath)

    def get_url(self, context=None, target=None):
        """Construct the url for the path."""
        url = None

        # See if a relative url is requested and get that if it is
        if context is not None and context.get('relative_links', True):
            url = self._get_relative_url(context, target)

        # If a relative url could not be produced or one was not wanted, get
        # an absolute url.
        if url is None:
            url = self._get_absolute_url(context, target)

        # Cleanup the url by  and double slashes
        url = url.rstrip('/')  # stripping leading slashes
        url = re_dbl_slash.sub('/', url)  # strip dble leading slash

        return url

    def _get_relative_url(self, context, target=None):
        """Construct the relative url for the path."""
        # Get the target_filepath. Prepend a '.' if needed. ex: '.html'
        target = str(target or self.target)
        target = '.' + target if not target.startswith('.') else target

        # Get the target_filepath and target_path. This should be from the
        # document that owns the this context or
        document = context.get('document', None)

        # If one isn't found, try using the root document. This can happen
        # if a mock context is used to simulate a document, like the toc
        # from the EpubBuilder
        document = document or context.get('root_document', None)

        # Dereference the document, if it's a weakref
        document = (document() if isinstance(document, weakref.ref) else
                    document)

        if document is not None:
            target_filepath = document.target_filepath(target)
            target_path = target_filepath.parent
        else:
            # Give up
            return None

        # Get the relative path
        relpath = os.path.relpath(self, target_path)

        # Construct a relative url relative to the target_path
        return str(relpath)

    def _get_absolute_url(self, context, target=None):
        """Construct the absolute url for the path."""
        # Get the parts of paths that will be used
        target_root = str(self.target_root).strip('.')
        target = str(target or self.target).strip('.')
        subpath = str(self.subpath).strip('.')

        url_str = (context['base_url'] if isinstance(context, dict) and
                   'base_url' in context else '/{target}/{subpath}')

        return url_str.format(target_root=target_root, target=target,
                              subpath=subpath)

    def use_name(self, name):
        """with_name(name) that returns a TargetPath.

        A new method name is used since the flavor_cls method takes
        precedence.

        Examples
        --------
        >>> p = TargetPath(target_root='/media', subpath='tests/fig1.png')
        >>> p.use_name('fig2.png')
        TargetPath('/media/tests/fig2.png')
        >>> p.use_name('fig2.png').subpath
        TargetPath('tests/fig2.png')
        """
        return TargetPath(target_root=self.target_root,
                          target=self.target,
                          subpath=self.subpath.with_name(name))

    def use_suffix(self, suffix):
        """with_suffix(suffix) that returns a TargetPath.

        A new method name is used since the flavor_cls method takes
        precedence.

        Examples
        --------
        >>> p = TargetPath(target_root='/media', subpath='tests/fig1.png')
        >>> p.use_suffix('.pdf')
        TargetPath('/media/tests/fig1.pdf')
        >>> p.use_suffix('.pdf').subpath
        TargetPath('tests/fig1.pdf')
        """
        return TargetPath(target_root=self.target_root,
                          target=self.target,
                          subpath=self.subpath.with_suffix(suffix))

    def use_subpath(self, subpath):
        """Returns a TargetPath with the subpath replaced with the given string
        or path.

        Examples
        --------
        >>> p = TargetPath(target_root='/media', subpath='tests/fig1.png')
        >>> p.use_subpath('text.txt')
        TargetPath('/media/text.txt')
        >>> p.use_subpath('text.txt').subpath
        TargetPath('text.txt')
        """
        return TargetPath(target_root=self.target_root,
                          target=self.target,
                          subpath=subpath)



"""
Builder for tex to pdf using latexmk
"""
from .builder import Builder


class Latexmk(Builder):
    """Compile a latex document into pdf using latexmk

    .. note:: This builder has been disabled by default because it fails about
              20% of the time with this system's ThreadPoolExecutor
              implementation.
    """

    action = ("latexmk "
              "-pdf "  # output pdf
              "-norc "  # do not use latexmkrc config
              "-f "  # continue past errors
              "-bibtex- "  # don't use bibtex
              "-jobname={builder.jobname} "  # filename of output file
              "-output-directory={builder.cache_path} "  # dir for temp files
              "{builder.infilepaths}")  # tex file to use

    available = False
    priority = 5000
    required_execs = ('latexmk', 'pdflatex')

    infilepath_ext = '.tex'
    outfilepath_ext = '.pdf'

    @property
    def cache_path(self):
        return self.outfilepath.parent

    @property
    def jobname(self):
        return self.outfilepath.stem



"""
A tag for a feature box used in some textbooks (like examples)
"""
from .tag import Tag


class FeatureBox(Tag):
    """A box with a feature for some texts, like an Example or Note box.

    Attributes
    ----------
    active : bool
        This tag is active.
    include_paragraphs : bool
        The contents of this tag cannot be included in paragraphs.
    html_classes : str
        Classes to append in the html tag.
    """

    active = True
    include_paragraphs = False

    html_name = "div"
    html_classes = "featurebox"
    tex_env = "featurebox"

    def __init__(self, name, *args, **kwargs):
        super().__init__(name='featurebox', *args, **kwargs)

    def html_fmt(self, attributes=None, **kwargs):
        attributes = (self.attributes.copy() if attributes is None else
                      attributes)

        # Add the html attribute
        attributes['class'] = (attributes['class'] + " " + self.html_classes
                               if 'class' in attributes else self.html_classes)
        return super().html_fmt(attributes=attributes, **kwargs)


class ExampleBox(FeatureBox):
    """A box for examples"""

    html_classes = "featurebox examplebox"
    tex_env = "examplebox"


class ProblemBox(FeatureBox):
    """A box for problems"""

    html_classes = "featurebox problembox"
    tex_env = "problembox"
